import numpy as np
import matplotlib.pyplot as plt
import cv2
import scipy.ndimage as ndimage
import scipy.ndimage.filters as filters
import math

def find_peaks(image=None, method='local-maximum', **kwargs):
    if method == 'AKAZE':
        coordinates = analyze(image, **kwargs)[2]
    elif method == 'absolute-threshold':
        coordinates = find_peaks_absolute_threshold(image, **kwargs)
    elif method == 'adaptive-threshold':
        coordinates = find_peaks_adaptive_threshold(image, **kwargs)
    elif method == 'local-maximum':
        coordinates = find_peaks_local_maximum(image, **kwargs)
    elif method == 'local-maximum-auto':
        coordinates = find_peaks_local_maximum_auto(image, **kwargs)
    elif method == 'relative-local-maximum':
        coordinates = find_peaks_relative_local_maximum(image, **kwargs)
    else:
        raise ValueError(f'Unkown method {method}.')
    return coordinates

def find_peaks_absolute_threshold(image, threshold = None, minimum_area = 5, maximum_area = 15):
    if threshold is None: threshold = (np.max(image) + np.min(image)) / 2
    image_thresholded = image > threshold
    coordinates = coordinates_from_contours(image_thresholded, minimum_area, maximum_area)
    return coordinates

def find_peaks_adaptive_threshold(image, minimum_area = 5, maximum_area = 15):
    # This may be needed if we go from a 16-bit image to an 8 bit image
    # if bounds is None:
    #     lower_bound = np.min(image)
    #     upper_bound = np.percentile(image.flatten(), 99.999)
    # else:
    #     lower_bound = bounds[0]
    #     upper_bound = bounds[1]
    #
    # # Change threshold and image to 8-bit, as cv2 can only analyse 8-bit images
    # # threshold = ((threshold - lower_bound) / (upper_bound - lower_bound) * 255)
    # # threshold = np.clip(threshold, 0, 255).astype('uint8')
    # image = ((image - lower_bound) / (upper_bound - lower_bound) * 255)
    # image = np.clip(image, 0, 255).astype('uint8')

    image_thresholded = cv2.adaptiveThreshold(image.astype(np.uint8),
                                              maxValue=1,
                                              adaptiveMethod=cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
                                              thresholdType=cv2.THRESH_BINARY,
                                              blockSize=9,
                                              C=0)
    coordinates = coordinates_from_contours(image_thresholded, minimum_area, maximum_area)

    return coordinates

def find_peaks_local_maximum(image,
                             minimum_intensity_difference=25,
                             maximum_intensity_difference=math.inf,
                             filter_neighbourhood_size_min=10,
                             filter_neighbourhood_size_max=5):

    """
    Find peaks based on local maxima.

    Parameters
    ----------
    image : NxM numpy.ndarray

    minimum_intensity_difference : int or float
        Lower threshold on the intensity difference between local minimum and local maximu
    maximum_intensity_difference : int or float, optional
        Upper threshold on the intensity difference between local minimum and local maximum.
    filter_neighbourhood_size_min : int
        Size of the minimum filter.
    filter_neighbourhood_size_max : int
        Size of the maximum filter.

    Returns
    -------
    Nx2 numpy.ndarray
        Peak coordinates

    Notes
    -----
    First minimum filtered and maximum filtered images are created from the input image.
    Local maxima are determined by finding the pixels where the maximum filtered image is equal to the input image.
    Local maxima where the intensity difference between the local maximum and local minimum is outside the interval
    (minimum_intensity_difference, maximum_intensity_difference) are discarded.
    """

    # Perhaps make filter round?
    image_max = filters.maximum_filter(image, filter_neighbourhood_size_max)
    maxima = (image == image_max)
    image_min = filters.minimum_filter(image, filter_neighbourhood_size_min)
    # Probably I need to make the neighbourhood_size of the minimum filter larger.

    difference_above_minimum = ((image_max - image_min) > minimum_intensity_difference)
    difference_below_maximum = ((image_max - image_min) < maximum_intensity_difference)
    difference_within_bounds = np.logical_and(difference_above_minimum, difference_below_maximum)
    maxima[difference_within_bounds == 0] = 0

    coordinates = np.fliplr(np.vstack(np.where(maxima)).T)

    # labeled, num_objects = ndimage.label(maxima)
    # if num_objects > 0:
    #     coordinates = np.fliplr(np.array(ndimage.center_of_mass(image, labeled, range(1, num_objects + 1))))
    # else:
    #     coordinates = np.array([])
    #     print('No peaks found')

    return coordinates



def find_peaks_local_maximum_auto(image,
                                  fraction_difference = 0.5,
                                  filter_neighbourhood_size_min=10,
                                  filter_neighbourhood_size_max=5):

    """
    Find peaks based on local maxima.

    Parameters
    ----------
    image : NxM numpy.ndarray

    minimum_intensity_difference : int or float
        Lower threshold on the intensity difference between local minimum and local maximu
    maximum_intensity_difference : int or float, optional
        Upper threshold on the intensity difference between local minimum and local maximum.
    filter_neighbourhood_size_min : int
        Size of the minimum filter.
    filter_neighbourhood_size_max : int
        Size of the maximum filter.

    Returns
    -------
    Nx2 numpy.ndarray
        Peak coordinates

    Notes
    -----
    First minimum filtered and maximum filtered images are created from the input image.
    Local maxima are determined by finding the pixels where the maximum filtered image is equal to the input image.
    Local maxima where the intensity difference between the local maximum and local minimum is outside the interval
    (minimum_intensity_difference, maximum_intensity_difference) are discarded.
    """

    # Perhaps make filter round?
    image_max = filters.maximum_filter(image, filter_neighbourhood_size_max)
    maxima = (image == image_max)
    image_min = filters.minimum_filter(image, filter_neighbourhood_size_min)
    # Probably I need to make the neighbourhood_size of the minimum filter larger.

    maxima_values = image_max[maxima]


    maxima_values = maxima_values[maxima_values>np.max(image_min)+np.std(image_min)*5]

    minimum_intensity_difference = (np.median(maxima_values) - np.median(image_min.flatten())) * fraction_difference

    difference_above_minimum = ((image_max - image_min) > minimum_intensity_difference)
    # difference_below_maximum = ((image_max - image_min) < maximum_intensity_difference)
    # difference_within_bounds = np.logical_and(difference_above_minimum, difference_below_maximum)
    difference_within_bounds = difference_above_minimum
    maxima[difference_within_bounds == 0] = 0

    coordinates = np.fliplr(np.vstack(np.where(maxima)).T)

    # labeled, num_objects = ndimage.label(maxima)
    # if num_objects > 0:
    #     coordinates = np.fliplr(np.array(ndimage.center_of_mass(image, labeled, range(1, num_objects + 1))))
    # else:
    #     coordinates = np.array([])
    #     print('No peaks found')

    return coordinates


def find_peaks_relative_local_maximum(image,
                                      minimum_times_background=4.5,
                                      maximum_times_background=math.inf,
                                      filter_neighbourhood_size_min=10,
                                      filter_sigma_min=25,
                                      filter_neighbourhood_size_max=5):

    image_min = filters.minimum_filter(image, filter_neighbourhood_size_min)
    image_min_gaussian = filters.gaussian_filter(image_min, sigma=filter_sigma_min)
    relative_image = image/image_min_gaussian

    relative_image_max = filters.maximum_filter(relative_image, filter_neighbourhood_size_max)
    maxima = (relative_image == relative_image_max)
    maxima[relative_image_max < minimum_times_background] = False

    labeled, num_objects = ndimage.label(maxima)
    if num_objects > 0:
        coordinates = np.fliplr(np.array(ndimage.center_of_mass(image, labeled, range(1, num_objects + 1))))
    else:
        coordinates = np.array([])
        print('No peaks found')

    return coordinates

def coordinates_from_contours(image_thresholded, minimum_area=5, maximum_area=15):
    contours, hierarchy = cv2.findContours(image_thresholded.astype(np.uint8),
                                           cv2.RETR_EXTERNAL,
                                           cv2.CHAIN_APPROX_NONE)
    x = []
    y = []

    # colorImg = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)

    coordinates = []

    for c in contours:
        # Calculate moments for each contour
        M = cv2.moments(c)

        # Calculate x, y coordinate of center

        if M["m00"] != 0:
            cX = int(M["m10"] / M["m00"])
            cY = int(M["m01"] / M["m00"])

            area = M["m00"]  # Is the same as cv2.contourArea(c) # Is the same as M["m00"]

            if (area > minimum_area) and (area < maximum_area):
                x = np.append(x, cX)
                y = np.append(y, cY)
                coordinates.append(np.array([cX, cY]))
        else:
            cX, cY = 0, 0

        # cv2.circle(colorImg, (cX, cY), 8, (0, 0, 255), thickness=1)

    return np.array(coordinates)


# # -*- coding: utf-8 -*-
# """
# Created on Wed Apr 17 13:50:42 2019
#
# @author: https://stackoverflow.com/questions/35854197/how-to-use-opencvs-connected-components-with-stats-in-python
#
# returns the number of spots, and per spot its centroid and the number of pixels (can be used to discard too large spots)
# """
# import cv2
# import bisect #This module provides support for maintaining a list in sorted order without having to sort the list after each insertion.
# import numpy as np
# import matplotlib.pyplot as plt
#
# from movie.background_correction import get_threshold
#
# def analyze(src, threshold=None):
#     if not threshold:
#         fL = get_threshold(src)
#     else:
#         fL = threshold
#     # gray1 = enhance_blobies_single(src, fL, 1)  # remove_background(src, fL)
#     gray1 = src
#     detector = cv2.AKAZE_create()
#     (kps1, descs1) = detector.detectAndCompute(gray1, None)
#     ctrd = np.array(cv2.KeyPoint_convert(kps1))
#
#     # remove all pixels at the edge (within 10 pix)
#     num_labels = len(ctrd)
#     dim1, dim0 = np.shape(src)
#     for ii in range(num_labels - 1, -1, -1):
#         discard = ctrd[ii, 0] < 10 or ctrd[ii, 1] < 10 or ctrd[ii, 0] > dim0 - 10 or ctrd[ii, 1] > dim1 - 10 or src[
#             int(ctrd[ii, 1]), int(ctrd[ii, 0])] == 0  # or size_label[ii]>100
#         # for some reason also spots are found on immean20 with no intensity --> discard
#         if discard:
#             ctrd = np.delete(ctrd, ii, axis=0)
#     #         size_label=np.delete(size_label,ii, axis=0)
#     num_labels = len(ctrd)
#
#     return num_labels, 0, ctrd
#
#
# def imadjust(src, tol=1, vout=(0, 255)):
#     # src : input one-layer image (numpy array)
#     # tol : tolerance, from 0 to 100.
#     # vin  : src image bounds
#     # vout : dst image bounds
#     # return : output img
#
#     assert len(src.shape) == 2, 'Input image should be 2-dims'
#
#     tol = max(0, min(100, tol))
#
#     vin = [np.min(src), np.max(src)]
#     vout = [0, 65535]  # 65535=16 bits
#     if tol > 0:
#         # Compute in and out limits
#         # Histogram
#         hist = np.histogram(src, bins=list(range(vin[1] - vin[0])), range=tuple(vin))[0]
#
#         # Cumulative histogram
#         cum = hist.copy()
#         for i in range(0, vin[1] - vin[0] - 1): cum[i] = cum[i - 1] + hist[i]  # why not hist.cumsum() here?
#
#         # Compute bounds
#         total = src.shape[0] * src.shape[1]
#         low_bound = total * tol / 100
#         upp_bound = total * (100 - tol) / 100
#         vin[0] = bisect.bisect_left(cum, low_bound)
#         vin[1] = bisect.bisect_left(cum, upp_bound)
#
#     # Stretching
#     scale = (vout[1] - vout[0]) / (vin[1] - vin[0])
#     vs = src - vin[0]
#     vs[src < vin[0]] = 0  # everything below zero becomes 0
#     vd = vs * scale + 0.5 + vout[0]  # why +0.5?
#     vd[vd > vout[1]] = vout[1]
#     dst = vd
#
#     return dst.astype(np.uint16)
#
#
# def im_binarize(img, f):
#     temp = img.copy()
#     temp[temp < f] = 0
#     return temp.astype(np.uint8)
#
#
# def enhance_blobies(image, f):
#     l, r = image[:, :image.shape[1] // 2], image[:, image.shape[1] // 2:]
#     l_adj, r_adj = imadjust(l.copy()), imadjust(r.copy())
#     l_bin, r_bin = im_binarize(l_adj, f).astype(np.uint8), im_binarize(r_adj, f).astype(np.uint8)
#     return l, r, l_bin, r_bin
#
# detector = cv2.AKAZE_create()
#
# (kps1, descs1) = detector.detectAndCompute(gray1, None);
# (kps2, descs2) = detector.detectAndCompute(gray2, None);
#
# print("keypoints: {}, descriptors: {}".format(len(kps1), descs1.shape))
# print("keypoints: {}, descriptors: {}".format(len(kps2), descs2.shape))
#
# # Match the features
# bf = cv2.BFMatcher(cv2.NORM_HAMMING)
# matches = bf.knnMatch(descs1, descs2, k=2)  # typo fixed
#
# # Apply ratio test
# pts1, pts2 = [], []
# for m in matches:
#     pts1.append(kps1[m[0].queryIdx].pt)
#     pts2.append(kps2[m[0].trainIdx].pt)
#
# pts1 = np.array(pts1).astype(np.float32)  # xy position
# pts2 = np.array(pts2).astype(np.float32)
# # AA=cv2.KeyPoint_convert(kps1);
#
# transformation_matrix, mask = cv2.findHomography(pts2, pts1, cv2.RANSAC, 20)
#


if __name__ == '__main__':
    from papylio import Experiment
    exp = Experiment(r'D:\ivoseverins\SURFdrive\Promotie\Code\Python\traceAnalysis\twoColourExampleData\20141017 - Holliday junction - Copy\HJC-50pM')
    file = exp.files[2]
    movie = file.movie
    image = movie.make_average_image(write=False)

    coordinates = find_peaks(image=image, method='adaptive-threshold', minimum_area=5, maximum_area=15)

    plt.imshow(image)
    plt.scatter(coordinates[:,0],coordinates[:,1],color='r')




import pytest
import xarray as xr
import tifffile
import numpy as np


@pytest.fixture
def experiment_output(shared_datadir):
    from papylio import Experiment
    return Experiment(shared_datadir / 'BN_TIRF_output_test_file')

@pytest.fixture
def files(experiment_output):
    return experiment_output.files

def test_file_collection(files):
    files

def test_get_variable(files):
    assert isinstance(files.serial.get_variable('intensity'), xr.DataArray)

def test_show_histogram(files):
    files.show_histogram('intensity')
    files.show_histogram('FRET', bins=100, range=(0,1))

#TODO: Make new test dataset that does not contain background which has illumination as first dimension.
# def test_merge_datasets(files):
#     files.merge_datasets()



import numpy as np
import xarray as xr
#
# stacking_energies = xr.DataArray(np.nan, dims=('reference', 'from_base', 'to_base'),
#                     coords={'reference': ['Protozanova2004', 'Punnoose2023'],
#                             'from_base': ['A','T','G','C'], 'to_base': ['A','T','G','C']})
#
#
# stacking_energies.loc[dict(reference='Protozanova2004')] = \
#     [[-1.11, -1.34, -1.06, -1.81],
#      [-0.19, -1.11, -0.55, -1.43],
#      [-1.43, -1.81, -1.44, -2.17],
#      [-0.55, -1.06, -0.91, -1.44]]
#
# stacking_energies.loc[dict(reference='Punnoose2023')] = \
#     [[-2.3, -1.5, -2.3, -1.9],
#      [-1.5, -0.8, -1.7, -0.5],
#      [-2.3, -1.7, -1.8, -2.1],
#      [-1.9, -0.5, -2.1, -0.6]]
#
# stacking_energies_flat = stacking_energies.stack(base_combination=('from_base', 'to_base'))
# stacking_energies_flat = stacking_energies_flat.reset_index('base_combination') \
#     .assign_coords(base_combination=('base_combination', [i+j for i,j in stacking_energies_flat.indexes['base_combination']]))


#
# def calculate_stacking_energy(sequence_subsets):
#
#     states = list(from_to_positions_per_state.keys())
#     energies = xr.DataArray(np.nan, dims=('sequence_subset', 'reference', 'state'),
#                             coords={'sequence_subset': sequence_subsets, 'state': states,
#                                     'reference': stacking_energies.reference})
#     for sequence_subset in sequence_subsets:
#         for state in (0,1):
#             from_tos = [(sequence_subset[i], sequence_subset[j]) for i, j in from_to_positions_per_state[state]]
#             energy = stacking_energies.stack(from_to=('from_base', 'to_base')).sel(from_to=from_tos).sum(dim='from_to')
#             energies.loc[dict(sequence_subset=sequence_subset, state=state)] = energy



def energy_dict_to_dataarray(energy_dict):
    return xr.DataArray(list(energy_dict.values()), dims=('base_combination',), coords={'base_combination': list(energy_dict.keys())})

stacking_energies_flat = xr.Dataset()

energy_dict = {'AA': -1.11, 'AT': -1.34, 'AG': -1.06, 'AC': -1.81,
               'TA': -0.19, 'TT': -1.11, 'TG': -0.55, 'TC': -1.43,
               'GA': -1.43, 'GT': -1.81, 'GG': -1.44, 'GC': -2.17,
               'CA': -0.55, 'CT': -1.06, 'CG': -0.91, 'CC': -1.44}
stacking_energies_flat['Protozanova2004'] = energy_dict_to_dataarray(energy_dict)

energy_dict = {'AA': -1.49, 'AT': -1.72, 'AG': -1.44, 'AC': -2.19,
               'TA': -0.57, 'TT': -1.49, 'TG': -0.93, 'TC': -1.81,
               'GA': -1.81, 'GT': -2.19, 'GG': -1.82, 'GC': -2.55,
               'CA': -0.93, 'CT': -1.44, 'CG': -1.29, 'CC': -1.82}
stacking_energies_flat['Krueger2006'] = energy_dict_to_dataarray(energy_dict)

energy_dict = {'AA': -1.36, 'AC': -2.03, 'AG': -1.60, 'AT': -2.35,
               'CA': -0.81, 'CC': -1.64, 'CG': -2.06, 'CT': -1.60,
               'GA': -1.39, 'GC': -3.42, 'GG': -1.64, 'GT': -2.03,
               'TA': -1.01, 'TC': -1.39, 'TG': -0.81, 'TT': -1.36}
stacking_energies_flat['Kilchherr2016'] = energy_dict_to_dataarray(energy_dict)


energy_dict = {'GA': -2.3, 'AG': -2.3, 'AA': -2.3, 'GG': -1.8, 'GC': -2.1, 'CG': -2.1, 'AC': -1.9, 'CA': -1.9,
               'GT': -1.7, 'TG': -1.7, 'AT': -1.5, 'TA': -1.5, 'TT': -0.8, 'CC': -0.6, 'CT': -0.5, 'TC': -0.5}
stacking_energies_flat['Punnoose2023'] = energy_dict_to_dataarray(energy_dict)

energy_dict = {'AA': -2.38, 'TT': -1.52, 'AT': -2.17, 'TA': -1.52, 'GG': -1.62, 'CC': -1.89, 'GC': -2.99, 'CG': -1.42,
                'TG': -1.42, 'CA': -1.45, 'AG': -1.74, 'CT': -1.19, 'TC': -2.19, 'GA': -2.16, 'GT': -2.06, 'AC': -3.57}
stacking_energies_flat['Banerjee2022'] = energy_dict_to_dataarray(energy_dict)

energy_dict = {'AA': -2.30, 'TT': -1.31, 'AT': -1.96, 'TA': -1.37, 'GG': -1.62, 'CC': -1.60, 'GC': -2.61, 'CG': -1.24,
                'TG': -1.26, 'CA': -1.29, 'AG': -1.71, 'CT': -0.95, 'TC': -1.92, 'GA': -2.00, 'GT': -1.80, 'AC': -3.22}
stacking_energies_flat['Banerjee2023'] = energy_dict_to_dataarray(energy_dict)

stacking_energies_flat = stacking_energies_flat.to_array('reference')

sequence_subsets = np.array(['TTAGCCGA', 'AATCGGCT', 'GGCGCCGC'])

from_to_positions_per_state = {0: ((0, 1), (2, 7), (6, 3), (4, 5)), 1: ((2, 3), (4, 1), (0, 5), (6, 7))}
states = [0,1]


def migrate_junction(sequence_subset, penultimate_bases='CCGCGGCG', step=-1):
    if step == -1: #Horizontal
        return penultimate_bases[0] + sequence_subset[0] + sequence_subset[3] + penultimate_bases[3] + \
               penultimate_bases[4] + sequence_subset[4] + sequence_subset[7] + penultimate_bases[7]
    elif step == 0:
        return sequence_subset
    elif step == 1: # Vertical
        return sequence_subset[1] + penultimate_bases[1] + penultimate_bases[2] + sequence_subset[2] + \
               sequence_subset[5] + penultimate_bases[5] + penultimate_bases[6] + sequence_subset[6]

base_pairs = ['AT','TA','CG','GC']
def check_basepairing(sequence_subset):
    if (sequence_subset[1:3] in base_pairs) and (sequence_subset[3:5] in base_pairs) and \
        (sequence_subset[5:7] in base_pairs) and ((sequence_subset[7] + sequence_subset[0]) in base_pairs):
        return True
    else:
        return False


def migration_options(sequence_subset, penultimate_bases='CCGCGGCG', return_all=False):
    migration_option_list = []
    for step in [-1, 0, 1]:
       sequence_subset_migrated = migrate_junction(sequence_subset, penultimate_bases=penultimate_bases, step=step)
       if check_basepairing(sequence_subset_migrated):
           migration_option_list.append(sequence_subset_migrated)
       elif return_all:
           migration_option_list.append(None)
    return migration_option_list

def migration_sequence_subsets(sequence_subsets, penultimate_bases='CCGCGGCGC'):
    migration_sequences = np.array([migration_options(sequence_subset, penultimate_bases=penultimate_bases, return_all=True)
                                    for sequence_subset in sequence_subsets])
    migration_sequences[migration_sequences == None] = ''
    migration_sequences = migration_sequences.astype('U')
    return migration_sequences

import itertools
bases = 'ATCG'
base_combinations = [''.join(i) for i in itertools.product(bases, repeat=2)]

def base_combination_per_position(sequence_subsets):
    base_combinations_per_position = xr.DataArray('', dims=('sequence_subset', 'position'),
                                          coords={'sequence_subset': sequence_subsets, 'position': np.arange(4)})

    for sequence_subset in sequence_subsets:
        for state in states:
            base_combinations_per_position.loc[dict(sequence_subset=sequence_subset, state=state)] = \
                [sequence_subset[i] + sequence_subset[j] for i, j in from_to_positions_per_state[state]]

def base_combination_count(sequence_subsets, penultimate_bases=None):
    base_combination_count = xr.DataArray(0, dims=('sequence_subset', 'base_combination', 'state','stack_position'),
                                          coords={'sequence_subset': sequence_subsets,
                                                  'base_combination': base_combinations,
                                                  'state': np.arange(2),
                                                  'stack_position': np.arange(4)})

    for sequence_subset in sequence_subsets:
        for state in states:
            for stack_position, positions in enumerate(from_to_positions_per_state[state]):
                # from_tos = [sequence_subset[i]+sequence_subset[j] for i, j in from_to_positions_per_state[state]]
                # base_combination, count = np.unique(from_tos, return_counts=True)
                # if state==0:
                #     count = -count
                base_combination = sequence_subset[positions[0]] + sequence_subset[positions[1]]
                base_combination_count.loc[dict(sequence_subset=sequence_subset,
                                                base_combination=base_combination,
                                                state=state, stack_position=stack_position)] += 1

                if penultimate_bases is not None:
                    penultimate_base_combination_0 = penultimate_bases[positions[0]] + sequence_subset[positions[0]]
                    penultimate_base_combination_1 = sequence_subset[positions[1]] + penultimate_bases[positions[1]]
                    base_combination_count.loc[dict(sequence_subset=sequence_subset,
                                                    base_combination=penultimate_base_combination_0,
                                                    state=state, stack_position=stack_position)] += 1
                    base_combination_count.loc[dict(sequence_subset=sequence_subset,
                                                    base_combination=penultimate_base_combination_1,
                                                    state=state, stack_position=stack_position)] += 1
    return base_combination_count



def calculate_total_stacking_energy(sequence_subsets, penultimate_bases=None):
    return (base_combination_count(sequence_subsets, penultimate_bases=penultimate_bases)
            .sum('stack_position').diff('state').squeeze(drop=True) * stacking_energies_flat).sum('base_combination')


# sequence_subsets = np.array(['TTAGCCGA', 'AATCGGCT', 'GGCGCCGC'])
# penultimate_bases = 'CCGCGGCGC'


def stacking_energies(sequence_subsets,penultimate_bases = 'CCGCGGCG'):
    return (base_combination_count(sequence_subsets, penultimate_bases=penultimate_bases)*stacking_energies_flat)

def total_stacking_energies(sequence_subsets, penultimate_bases = 'CCGCGGCG'):
    return stacking_energies(sequence_subsets, penultimate_bases).sum('base_combination').sum('stack_position')

def inner_stacking_energies(sequence_subsets, penultimate_bases = 'CCGCGGCG'):
    return stacking_energies(sequence_subsets, penultimate_bases).sel(stack_position=[1,2]).sum('base_combination').sum('stack_position')

def inner_stacking_energies_minimum(sequence_subsets, penultimate_bases='CCGCGGCG'):
    return stacking_energies(sequence_subsets, penultimate_bases).sel(stack_position=[1, 2]).sum('base_combination').min('stack_position')

    #
    #
    #
    # sequence_subset_split = np.array(list(sequence_subset))
    # if state == 0:
    #     sequence_subset_split = np.array(list(sequence_subset))
    #     from_bases = sequence_subset_split[[0,2,6,4]]
    #     to_bases = sequence_subset_split[[1,7,3,5]]
    #
    #     from_tos = [(sequence_subset[i], sequence_subset[j]) for i,j in ((0,1),(2,7),(6,3),(4,5))]
    #
    #
    #
    # if state == 1:
    #     from_bases = sequence_subset_split[[2,4,0,6]]
    #     to_bases = sequence_subset_split[[3,1,5,7]]
    # # energies_from_ref = stacking_energies.sel(reference=)
    # # np.array([stacking_energies.sel(from_base=from_base, to_base=to_base) for from_base, to_base in zip(from_bases, to_bases)])
    # #
    # from_tos = np.stack([from_bases, to_bases]).T
    # energy_for_subset = stacking_energies.stack(from_to=('from_base', 'to_base')).sel(from_to=[tuple(ft) for ft in from_tos])
    # return energy_for_subset.sum(dim='from_to')



import numpy as np
import xarray as xr


def trace_classification_threshold(traces, threshold):
    classification_lower = classification_upper = True
    if threshold[0] is not None:
        classification_lower = np.vstack([(trace > threshold[0]) for trace in traces.values])
    if threshold[1] is not None:
        classification_upper = np.vstack([(trace < threshold[1]) for trace in traces.values])

    classification = xr.DataArray((classification_upper & classification_lower),
                                  dims=('molecule', 'frame'), name='classification')
    return classification


def trace_selection_threshold(traces, threshold):
    classification = trace_classification_threshold(traces, threshold)
    return classification.all(dim='frame')



def rolling_correlation(traces, rolling_dim='frame', correlation_dim='channel', window=10):
    windows = traces.rolling(dim={rolling_dim: window}, center=True, min_periods=1).construct(window_dim='section', stride=1, keep_attrs=None)

    mean_windows = windows.mean('section')
    windows_minus_mean = windows-mean_windows

    a = windows_minus_mean.prod(correlation_dim, skipna=False).sum('section')
    b = (windows_minus_mean**2).sum('section').prod(correlation_dim)**(1/2)
    p = a/b

    return p

def classify_correlation(traces, rolling_dim='frame', correlation_dim='channel', window=10, rolling_mean_window=10, threshold=0.75):
    rc = rolling_correlation(traces, rolling_dim=rolling_dim, correlation_dim=correlation_dim, window=window)
    rcm = rc.rolling(dim={rolling_dim: rolling_mean_window}, center=True, min_periods=1).mean()
    classification = (rcm > threshold).astype(int).rolling(dim={rolling_dim: rolling_mean_window}, center=True, min_periods=1).max()
    classification.name = 'classification'
    return classification


def classify_anticorrelation(traces, rolling_dim='frame', correlation_dim='channel', window=10, rolling_mean_window=10, threshold=-0.75):
    rc = rolling_correlation(traces, rolling_dim=rolling_dim, correlation_dim=correlation_dim, window=window)
    rcm = rc.rolling(dim={rolling_dim: rolling_mean_window}, center=True, min_periods=1).mean() # To smooth out variations
    classification = (rcm < threshold).astype(int).rolling(dim={rolling_dim: rolling_mean_window}, center=True, min_periods=1).max() # To widen the window
    classification.name = 'classification'
    return classification


import os
import numpy as np

from papylio.movie.movie import Movie


class PmaMovie(Movie):
    extensions = ['.pma']

    def __init__(self, arg, *args, **kwargs):
        super().__init__(arg, *args, **kwargs)
        
        self.writepath = self.filepath.parent
        self.name = self.filepath.with_suffix('').name

        self.channel_arrangement = np.array([[[0,1]]])

        # Determine whether the image is 8 bits or 16 bits
        if (self.filepath.name[-7:-4]=='_16'):
            self.data_type = np.dtype(np.uint16)
        else:
            self.data_type = np.dtype(np.uint8)

        # Is this still used? [IS: 20-04-2021]
        self.threshold = {  'view':             (0,200),
                            'point-selection':  (45,25)
                            }

        # self.read_header()
        # self.create_frame_info()  # Possibly move to Movie later on



    def open(self):
        pass  # TODO: implement this

    def close(self):
        pass  # TODO: implement this

    def _read_header(self):
        statinfo = os.stat(self.filepath)       
               
        with self.filepath.open('rb') as pma_file:
            self.width = np.fromfile(pma_file, np.int16, count=1)[0].astype(int)
            self.height = np.fromfile(pma_file, np.int16, count=1)[0].astype(int)
            self.number_of_frames = int((statinfo.st_size-4)/(self.width*self.height))

        # TODO: Import log file
        # self.exposure_time = np.genfromtxt(f'{self.absoluteFilePath}.log', max_rows=1)[2]
        # print(f'Exposure time set to {self.exposure_time} sec for {self.name}')
        # self.log_details = open(f'{self.absoluteFilePath}.log').readlines()
        # self.log_details = ''.join(self.log_details)

    def _read_frame(self, frame_number):
        with self.filepath.open('rb') as pma_file:
            np.fromfile(pma_file, np.uint16, count=1)
            np.fromfile(pma_file, np.uint16, count=1)
        
            if self.bitdepth == 8:
                pma_file.seek(4+(frame_number*(self.width*self.height)), os.SEEK_SET)
                image = np.reshape(np.fromfile(pma_file, np.uint8, count=self.width*self.height), (self.width,self.height))
            else:
                pma_file.seek(4+2*frame_number*(self.width*self.height), os.SEEK_SET)
                msb = np.reshape(np.fromfile(pma_file, np.uint8, count=(self.width*self.height)), (self.width, self.height))
                lsb = np.reshape(np.fromfile(pma_file, np.uint8, count=(self.width*self.height)), (self.width, self.height))
                image = 256*msb+lsb

        return image

    def _read_frames(self, indices):
        # Can probably be implemented more efficiently
        return np.stack([self._read_frame(i) for i in indices])

if __name__ == "__main__":
    movie = PmaMovie(r'.\Example_data\pma\movie.pma')
    # movie.intensity_range = (0, 120)
    # movie.make_projection_images()



# -*- coding: utf-8 -*-
"""
Created on Fri Dec  6 18:59:11 2019

@author: pimam
"""
if __name__ == '__main__':
    import os
    import sys
    from pathlib import Path, PureWindowsPath
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    p = Path(__file__).parents[3]
    sys.path.insert(0, str(p))
    main_path = PureWindowsPath('C:\\Users\\pimam\\Documents\\MEP\\tracesfiles')

import numpy as np
import pandas as pd
import matplotlib.pylab as plt
import seaborn as sns
sns.set(style="dark")
sns.set_color_codes()


def ML1expcut(dwells, Tcut, Ncut):
    if Ncut == 0:
        MLtau = np.average(dwells)
    else:
        Nrec = dwells.size
        avg_dwells = np.average(dwells)
        MLtau = avg_dwells + Ncut*Tcut/Nrec
    timearray = np.linspace(0, Tcut, 1000)
    P = 1/MLtau*np.exp(-timearray/MLtau)
    return P, MLtau


def P2expcut(dwells, params, Tcut, Ncut):
    P1, tau1, tau2 = params
    Pi = P1/tau1*np.exp(-dwells/tau1)+(1-P1)/tau2*np.exp(-dwells/tau2)
    Pcut = P1*np.exp(-Tcut/tau1)+(1-P1)*np.exp(-Tcut/tau2)
    return Pi, Pcut


def LogLikelihood(xdata, params, model, Tcut, Ncut):
    Pi, Pcut = model(xdata, params, Tcut, Ncut)
    LLikecut = 0
    if Ncut != 0:
        LLikecut = -Ncut * np.log(Pcut)
    LLike = np.sum(-np.log(Pi)) + LLikecut
    return LLike


def update_temp(T, alpha):
    '''
    Exponential cooling scheme
    :param T: current temperature.
    :param alpha: Cooling rate.
    :return: new temperature
    '''
    T *= alpha
    return T


def simmulated_annealing(data, objective_function, model, x_initial, lwrbnd,
                         uprbnd, Tcut, Ncut, Tstart=100.,
                         Tfinal=0.001, delta1=0.1, delta2=2.5, alpha=0.9):
    i = 0
    T = Tstart
    step = 0
    xstep = 0
    x = x_initial
    while T > Tfinal:
        step += 1
        if (step % 100 == 0):
            T = update_temp(T, alpha)
        x_trial = np.zeros(len(x))
        x_trial[0] = np.random.uniform(np.max([x[0] - delta1, lwrbnd[0]]),
                                       np.min([x[0] + delta1, uprbnd[0]]))
        for i in range(1, len(x)):
            x_trial[i] = np.random.uniform(np.max([x[i] - delta2, lwrbnd[i]]),
                                           np.min([x[i] + delta2, uprbnd[i]]))
        x, xstep = Metropolis(objective_function, model, x, x_trial, T, data,
                              Tcut, Ncut, xstep)
    print(f'steps: {xstep}')
    return x, xstep


def Metropolis(f, model, x, x_trial, T, data, Tcut, Ncut, xstep):
    # Metropolis Algorithm to decide if you accept the trial solution.
    Vnew = f(data, x_trial, model, Tcut, Ncut)
    Vold = f(data, x, model, Tcut, Ncut)
    if (np.random.uniform() < np.exp(-(Vnew - Vold) / T)):
        x = x_trial
        xstep += 1
    return x, xstep


def Best_of_Nfits_sim_anneal(dwells, Nfits, model, x_initial,
                             lwrbnd, uprbnd, Tcut, Ncut):
    # Perform N fits on data using simmulated annealing
    LLike = np.empty(Nfits)
    for i in range(0, Nfits):
        fitdata, xstep = simmulated_annealing(data=dwells,
                                              objective_function=LogLikelihood,
                                              model=model, x_initial=x_initial,
                                              lwrbnd=lwrbnd, uprbnd=uprbnd,
                                              Tcut=Tcut, Ncut=Ncut)
        print(f"fit{i} found: {fitdata}")
        if i == 0:
            fitparam = [fitdata]
            Nsteps = [xstep]
        else:
            fitparam = np.concatenate((fitparam, [fitdata]), axis=0)
            Nsteps = np.concatenate((Nsteps, [xstep]), axis=0)
        LLike[i] = LogLikelihood(dwells, fitparam[i], model, Tcut, Ncut)
    ibestparam = np.argmax(LLike)
    bestparam = fitparam[ibestparam]
    bestNsteps = Nsteps[ibestparam]
    return bestparam, bestNsteps

def Bootstrap_data(dwells, Ncut, Ntrial):
    dwells_Ncut = np.concatenate((dwells, np.zeros(Ncut)))
    dwells_rand = np.random.choice(dwells_Ncut, Ntrial)
    Bootstrapped_dwells = dwells_rand[dwells_rand > 0]
    Bootstrapped_Ncut = np.count_nonzero(dwells_rand == 0)
    return Bootstrapped_dwells, Bootstrapped_Ncut


def fit(dwells_all, mdl, dataset_name, Nfits=1, include_over_Tmax=True,
            bootstrap=False, boot_repeats=0):
    Tmax = dwells_all.max()
    if include_over_Tmax is True:
        Tcut = Tmax - 5
        dwells = dwells_all[dwells_all < Tcut]
        Ncut = dwells_all[dwells_all >= Tcut].size
    else:
        Tcut = 0
        Ncut = 0
        dwells = dwells_all
    print(f'Ncut: {Ncut}')

    fit_result = pd.DataFrame(columns=['Dataset', 'model', 'params',
                                       'values', 'error', 'Ncut', 'BootRepeats',
                                       'steps', 'chi-square', 'BIC'])

    if mdl == '1Exp':
        model = ML1expcut
        if bootstrap is True:
            Ntrial = 1000
            Ncutarray = np.empty(boot_repeats+1)
            Nstepsarray = np.full(boot_repeats+1, np.nan)
            for i in range(0, boot_repeats):
                boot_dwells, boot_Ncut = Bootstrap_data(dwells, Ncut, Ntrial)
                fit, param = model(boot_dwells, Tcut, boot_Ncut)
                Ncutarray[i] = boot_Ncut
                if i == 0:
                    params = [1, param, np.nan]
                    fitparam = [params]
                else:
                    params = [1, param, np.nan]
                    fitparam = np.concatenate((fitparam, [params]), axis=0)

            # Save data of interest to dataframe
            bestfit, bestparam = model(dwells, Tcut, Ncut)
            bestparams = [1, bestparam, np.nan]
            Allfitparam = np.concatenate((fitparam, [bestparams]), axis=0)
            data = pd.DataFrame(Allfitparam)
            # print("All fitparam: ", data)
            data.columns = ['P1', 'tau1', 'tau2']
            data['Nsteps'] = Nstepsarray
            data['Ncut'] = Ncutarray
            idx = []
            for i in range(len(fitparam)):
                idx.append('fit' + str(i+1))
            idx.append('Bestfit')
            data.index = idx
        else:
            fit, fitparam = ML1expcut(dwells, Tcut, Ncut)
            # Save data of interest to dataframe
            data = pd.DataFrame({'P1': [1], 'tau1': [fitparam],
                                 'tau2': [np.nan],
                                 'Nsteps': [np.nan], 'Ncut': [Ncut]})

    elif mdl == '2Exp':
        model = P2expcut

        # Set parameters for simmulated annealing
        avg_dwells = np.average(dwells)
        x_initial = [0.5, avg_dwells, avg_dwells]
        lwrbnd = [0, 0, 0]
        uprbnd = [1, 2*Tmax, 2*Tmax]

        # Check if bootstrapping is used
        if bootstrap is True:
            Ntrial = 1000
            LLike = np.empty(boot_repeats)
            Ncutarray = np.empty(boot_repeats)
            Nstepsarray = np.full(boot_repeats, np.nan)
            print('bootrepeats: ', boot_repeats)
            for i in range(0, boot_repeats):
                boot_dwells, boot_Ncut = Bootstrap_data(dwells, Ncut, Ntrial)
                param, Nsteps = Best_of_Nfits_sim_anneal(
                                                       boot_dwells, Nfits,
                                                       model=model,
                                                       x_initial=x_initial,
                                                       lwrbnd=lwrbnd,
                                                       uprbnd=uprbnd,
                                                       Tcut=Tcut,
                                                       Ncut=boot_Ncut)
                Ncutarray[i] = boot_Ncut
                Nstepsarray[i] = Nsteps
                if i == 0:
                    fitparam = [param]
                else:
                    fitparam = np.concatenate((fitparam, [param]), axis=0)
                LLike[i] = LogLikelihood(dwells, fitparam[i], model, Tcut, Ncut)
            ibestparam = np.argmax(LLike)

            # Save data of interest to dataframe
            bestparam = fitparam[ibestparam]
            bestNsteps = Nstepsarray[ibestparam]
            bestNcut = Ncutarray[ibestparam]
            Allfitparam = np.concatenate((fitparam, [bestparam]), axis=0)
            Nstepsarray = np.concatenate((Nstepsarray, [bestNsteps]), axis=0)
            Ncutarray = np.concatenate((Ncutarray, [bestNcut]), axis=0)
            data = pd.DataFrame(Allfitparam)
            data.columns = ['P1', 'tau1', 'tau2']
            data['Nsteps'] = Nstepsarray
            data['Ncut'] = Ncutarray
            idx = []
            for i in range(len(fitparam)):
                idx.append('fit' + str(i+1))
            idx.append('Bestfit')
            data.index = idx

        else:
            # Perform N fits on data using simmulated annealing and select best
            bestparam, bestNsteps = Best_of_Nfits_sim_anneal(
                                                       dwells, Nfits,
                                                       model=model,
                                                       x_initial=x_initial,
                                                       lwrbnd=lwrbnd,
                                                       uprbnd=uprbnd,
                                                       Tcut=Tcut,
                                                       Ncut=Ncut)
            data = pd.DataFrame({'P1': [bestparam[0]], 'tau1': [bestparam[1]],
                                'tau2': [bestparam[2]], 'Nsteps': [bestNsteps],
                                 'Ncut': [Ncut]})

    return data


if __name__ == '__main__':

    # Import data and prepare for fitting
    filename = '2exp_N=10000_rep=1_tau1=10_tau2=200_a=0.5'
    dwells_all = np.load('./data/2exp_N=10000_rep=1_tau1=10_tau2=200_a=0.5.npy')
    dwells_all = dwells_all[0]

    # Start fitting
    mdl = '2Exp'
    include_over_Tmax = True
    Nfits = 200
    bootstrap = True
    boot_repeats = 200
    fitdata = fit(dwells_all, mdl, Nfits, include_over_Tmax, bootstrap, boot_repeats)
    print(fitdata)
    if bootstrap is True:
        fitdata.to_csv(f'{mdl}_inclTmax_{include_over_Tmax}_bootstrap{boot_repeats}.csv', index=False)
    else:
        fitdata.to_csv(f'{mdl}_inclTmax_{include_over_Tmax}_Nfits{Nfits}.csv', index=False)

#    newdata = pd.read_csv(f'{mdl}_inclTmax_{include_over_Tmax}_bootstrap{boot_repeats}.csv')

    # Getting measures and plotting the parameter values found
    taubnd = 100
    fitP1 = []
    fittau1 = []
    fittau2 = []
    for i in range(0, len(fitdata['tau1'])):
        if fitdata['tau1'][i] > taubnd:
            fittau2.append(fitdata['tau1'][i])
            fitP1.append(1-fitdata['P1'][i])
        else:
            fittau1.append(fitdata['tau1'][i])
            fitP1.append(fitdata['P1'][i])
        if fitdata['tau2'][i] > taubnd:
            fittau2.append(fitdata['tau2'][i])
        else:
            fittau1.append(fitdata['tau2'][i])

    P1_avg = np.average(fitP1)
    tau1_avg = np.average(fittau1)
    tau2_avg = np.average(fittau2)
    P1_std = np.std(fitP1)
    tau1_std = np.std(fittau1)
    tau2_std = np.std(fittau2)
    Nbins = 50

    plt.figure()
    plt.hist(fitP1, bins=Nbins)
    plt.vlines(P1_avg, 0, round(Nbins/2), label='avg:'+"{0:.2f}".format(P1_avg))
    plt.title(f'Fit values for P1 Nfits: {boot_repeats} Nbins: {Nbins}')
    plt.legend()
    plt.figure()
    plt.hist(fittau1, bins=Nbins)
    plt.vlines(tau1_avg, 0, round(Nbins/2), label='avg:'+"{0:.2f}".format(tau1_avg))
    plt.title(rf'Fit values for $\tau$1 Nfits: {boot_repeats} Nbins: {Nbins}')
    plt.legend()
    plt.figure()
    plt.hist(fittau2, bins=Nbins)
    plt.vlines(tau2_avg, 0, round(Nbins/2), label='avg:'+"{0:.2f}".format(tau2_avg))
    plt.title(rf'Fit values for $\tau$1 Nfits: {boot_repeats} Nbins: {Nbins}')
    plt.legend()


#    # Plot data with double and single exponential fit
#    plt.figure()
#    plt.semilogy(centers, values, '.', label=f'Dwells with Ncut:{Ncut}')
#    plt.semilogy(timearray, bestfit, label='P1:'+"{0:.2f}".format(bestparams[0])+"\n"+r'$\tau$1:'+"{0:.1f}".format(bestparams[1])+"\n"+r'$\tau$2:'+"{0:.1f}".format(bestparams[2]))
#    singlexp = 1/avg_dwells*np.exp(-timearray/avg_dwells)
#    plt.plot(timearray, singlexp, 'orange', label = rf'$\tau$:{avg_dwells:.1f}')
#    plt.xlabel('dwell time (sec)')
#    plt.ylabel('log prob. density')
#    plt.legend(fontsize='x-large')
#  #  plt.savefig(f'{len(exp.files)}files_1_2expfit__compared.png', dpi=200)


# def ML2expcut(dwells, params, Tcut, Ncut):  # not used
#     P1, tau1, tau2 = params
#     Pi = P1/tau1*np.exp(-dwells/tau1)+(1-P1)/tau2*np.exp(-dwells/tau2)
#     LLike = np.sum(-np.log(Pi))
#     if Ncut != 0:
#         Pcut = P1*np.exp(-Tcut/tau1)+(1-P1)*np.exp(-Tcut/tau2)
#         LLikecut = -Ncut * np.log(Pcut)
#         LLike += LLikecut
#     return Pi, LLike


import sys

def return_none_when_executed_by_pycharm(func):
    def wrapper(*args, **kwargs):
        # print(sys._getframe(1))
        if sys._getframe(1).f_code.co_name=='generate_imports_tip_for_module':
            # print('return None')
            return None
        else:
            # print('return normal')
            return func(*args, **kwargs)
    return wrapper


import xarray as xr
import numpy as np
from itertools import accumulate, groupby
from hmmlearn import hmm
import pomegranate as pg
import tqdm
from objectlist import ObjectList
from copy import deepcopy
import scipy.linalg

# file.FRET, file.classification, file.selected
def hidden_markov_modelling(traces, classification, selection, n_states=2, threshold_state_mean=None, level='molecule'):
    if level == 'molecule':
        models_per_molecule = fit_hmm_to_individual_traces(traces, classification, selection, parallel=False, n_states=n_states, threshold_state_mean=threshold_state_mean)
    elif level == 'file':
        model = fit_hmm_to_file(traces, classification, selection, n_states=n_states, threshold_state_mean=threshold_state_mean)
        number_of_molecules = np.shape(traces)[0]
        models_per_molecule = [deepcopy(model) for _ in range(number_of_molecules)]
    else:
        raise RuntimeError('Hidden markov modelling can be performed on the molecule of file level. Indicate this with level=\'molecule\' or level=\'file\'')

    ds = xr.Dataset()
    if models_per_molecule is None:
        # TODO: Is this if statement still necessary, now that we set the return_none_if_all_none to False
        # ds['selection_complex_rates'] = xr.ones_like(selection, dtype=bool)
        # ds['selection_lower_rate_limit'] = xr.ones_like(selection, dtype=bool)
        # ds['classification_hmm'] = -xr.ones_like(classification)
        # return ds
        raise RuntimeError('If you see this error please let Ivo know.')

    ds['number_of_states'] = number_of_states_from_models(models_per_molecule)
    state_parameters = state_parameters_from_models(models_per_molecule, n_states=n_states)
    transition_matrices = transition_matrices_from_models(models_per_molecule, n_states=n_states)
    classification_hmm = trace_classification_models(traces, classification, models_per_molecule).astype('int8')
    # ds['state_parameters'], ds['transition_matrix'] = \
    state_parameters, transition_matrices, classification_hmm = \
        sort_states_in_data(state_parameters, transition_matrices, classification_hmm)
    ds['classification_hmm'] = classification_hmm

    ds['state_mean'] = state_parameters.sel(parameter=0)
    ds['state_standard_deviation'] = state_parameters.sel(parameter=1)
    ds['transition_probability'] = transition_matrices[:, :n_states, :n_states]
    ds['start_probability'] = transition_matrices[:, -2, :n_states]
    ds['end_probability'] = transition_matrices[:, :n_states, -1]


    number_of_frames = len(traces.frame)
    frame_rate = 1 / traces.time.diff('frame').mean().item()
    transition_rates = determine_transition_rates_from_probabilities(ds.number_of_states, ds.transition_probability,
                                                                     frame_rate)
    transition_rates, ds['selection_complex_rates'] = complex_transition_rates_to_nan(transition_rates)

    ds['transition_rate'], ds['selection_lower_rate_limit'] = \
        transition_rates_outside_measurement_resolution_to_nan(transition_rates, number_of_frames, frame_rate)

    return ds



def BIC(model, xis):
    if isinstance(model, pg.NormalDistribution):
        k = 2
        likelihood = model.probability(np.concatenate(xis)).prod()
    else:
        state_count = (model.state_count() - 2)
        k = state_count * (state_count - 1) + state_count * 2
        # likelihood = model.predict_proba(xi).max(axis=1).prod()
        log_likelihood = np.sum([model.log_probability(xii) for xii in xis])

    n = len(np.concatenate(xis))
    bic_value = -2 * log_likelihood + k * np.log(n)
    return bic_value

    #     likelihood = np.prod([model.probability(xii) for xii in xis])
    #
    # n = len(np.concatenate(xis))
    # bic_value_old = -2*np.log(likelihood) + k * np.log(n)


def split_by_classification(xi, classification):
    split_indices = np.nonzero(np.diff(classification))[0] + 1
    cis = np.split(classification, split_indices)
    xis = np.split(xi, split_indices)
    # cis = [cii[0] for cii in cis]
    return xis, cis


def hmm1and2(input):
    xi, classification, selected = input
    # xi = molecule.FRET.values
    # classification = molecule.classification.values
    if not selected:
        return None

    classification_st_0 = classification < 0
    if classification_st_0.all():
        return None
    if (~classification_st_0).sum() < 2:
        return None

    included_frame_selection = classification >= 0
    xis, cis = split_by_classification(xi, included_frame_selection)

    xis = [xii for cii, xii in zip(cis, xis) if cii[0]]

    dist1 = pg.NormalDistribution.from_samples(np.concatenate(xis))
    model1 = pg.HiddenMarkovModel.from_matrix([[1]], [dist1], [1])
    # model1 = pg.HiddenMarkovModel.from_samples(pg.NormalDistribution, n_components=1, X=[xi])
    model2 = pg.HiddenMarkovModel.from_samples(pg.NormalDistribution, n_components=2, X=xis)

    bic_model1 = BIC(model1, xis)
    bic_model2 = BIC(model2, xis)

    if bic_model1 < bic_model2:
        # parameters = np.array(model1.parameters)
        # transition_matrix = np.array([1])
        return model1
    elif bic_model2 < bic_model1:
        # parameters = np.vstack([model2.states[0].distribution.parameters, model2.states[1].distribution.parameters])
        # transition_matrix = model2.dense_transition_matrix()
        return model2
    else:
        # parameters = None
        # transition_matrix = None
        return None
        # raise(ValueError)

    # return parameters, transition_matrix

def hmm_n_states(input, n_states=2, threshold_state_mean=None, level='molecule'):

    xi, classification, selected = input

    if np.sum(selected) == 0:
        return None

    classification_st_0 = classification < 0
    if classification_st_0.all():
        return None
    if (~classification_st_0).sum() < 2:
        return None

    included_frame_selection = classification >= 0
    xis, cis = split_by_classification(xi, included_frame_selection)

    if level == 'molecule':
        xis = [xii for cii, xii in zip(cis, xis) if cii[0]]
    elif level == 'file':
        xis_new = []
        for xii, cii in zip(xis, cis):
            if len(xii[cii]) > 0:
                xis_new.append(xii[cii])
        xis = xis_new
    else:
        raise RuntimeError('Hidden markov modelling can be performed on the molecule of file level. Indicate this with level=\'molecule\' or level=\'file\'')

    best_model = None
    best_bic = np.inf

    for state in range(1, n_states + 1):
        if state == 1:
            dist1 = pg.NormalDistribution.from_samples(np.concatenate(xis))
            model = pg.HiddenMarkovModel.from_matrix([[1]], [dist1], [1])
        else:
            try:
                model = pg.HiddenMarkovModel.from_samples(pg.NormalDistribution, n_components=state, X=xis)
            except ValueError:
                continue

        bic = BIC(model, xis)

        if threshold_state_mean:
            state_means = []
            for state in model.states:
                if state.distribution:
                    if isinstance(state.distribution, pg.NormalDistribution):
                        state_means.append(state.distribution.parameters[0])

            def check_difference(state_means, threshold=threshold_state_mean):
                for i in range(len(state_means)):
                    for j in range(i + 1, len(state_means)):
                        if abs(state_means[i] - state_means[j]) < threshold:
                            return False
                return True

            result = check_difference(state_means, threshold_state_mean)

            if bic < best_bic and result:
                best_bic = bic
                best_model = model
        else:
            if bic < best_bic:
                best_bic = bic
                best_model = model

    if best_model is not None:
        return best_model
    else:
        return None

def fit_hmm_to_individual_traces(traces, classification, selected, parallel=False, n_states=2, threshold_state_mean=None):
    cf = ObjectList(list(zip(traces.values, classification.values, selected.values)), return_none_if_all_none=False)
    cf.use_parallel_processing = parallel
    models_per_molecule = cf.map(hmm_n_states)(n_states=n_states, threshold_state_mean=threshold_state_mean)  # New taking sections into account 5540 traces: 5:02
        # Old not taking sections into account: 5092 traces 4:00 minutes (2:37 on server)
    return models_per_molecule

def fit_hmm_to_file(traces, classification, selected, n_states=2, threshold_state_mean=None):
    input_values = [traces.values, classification.values, selected.values]
    models = hmm_n_states(input_values, n_states=n_states, threshold_state_mean=threshold_state_mean, level='file')
    return models

def number_of_states_from_models(models):

    number_of_states = [model.state_count()-2 if model is not None else 0 for model in models]
    # state_count = []
    # for model in models:
    #     if model is None:
    #         state_count.append(0)
    #     else:
    #         state_count.append(model.state_count()-2)

    return xr.DataArray(number_of_states, dims='molecule')

def state_parameters_from_models(models, n_states=2):
    max_number_of_states = n_states
    number_of_parameters = 2

    state_parameters = np.full((len(models), max_number_of_states, number_of_parameters), np.nan)
    for i, model in enumerate(models):
        if model is not None:
            sp = np.vstack([state.distribution.parameters for state in model.states[:-2]])
            state_parameters[i, :sp.shape[0], :] = sp
    return xr.DataArray(state_parameters, dims=('molecule', 'state', 'parameter'))

def transition_matrices_from_models(models, n_states=2):
    max_number_of_states = n_states
    transition_matrix = np.full((len(models), max_number_of_states+2, max_number_of_states+2), np.nan)
    for i, model in enumerate(models):
        if model is not None:
            tm = model.dense_transition_matrix()
            number_of_states = tm.shape[0]-2
            transition_matrix[i, :number_of_states, :number_of_states] = tm[:number_of_states, :number_of_states]
            transition_matrix[i, -2:, :number_of_states] = tm[-2:, :number_of_states]
            transition_matrix[i, :number_of_states, -2:] = tm[:number_of_states, -2:,]
            transition_matrix[i, -2:, -2:] = tm[-2:, -2:]
    #         start_index = max_number_of_states+2-tm.shape[0]
    #         transition_matrix[i,start_index:,start_index:] = tm
    # transition_matrix = transition_matrix[:, [1,0,2,3],:][:,:,[1,0,2,3]] # In case of a single state this put the state at index 0.
    return xr.DataArray(transition_matrix, dims=('molecule','from_state','to_state'))

def sort_states_in_data(state_parameters, transition_matrices, classification_hmm):
    sort_indices = state_parameters[:, :, 0].argsort(axis=1)

    sort_indices_start_end_states = xr.DataArray([[2, 3]] * len(state_parameters.molecule), dims=('molecule', 'state'))
    sort_indices_transition_matrix = xr.concat([sort_indices, sort_indices_start_end_states], dim='state')

    state_parameters = state_parameters.sel(state=xr.DataArray(sort_indices, dims=('molecule', 'state')))
    transition_matrices = transition_matrices.sel(from_state=sort_indices_transition_matrix.rename(state='from_state'),
                                              to_state=sort_indices_transition_matrix.rename(state='to_state'))

    classification_hmm_sorted = xr.DataArray(np.zeros_like(classification_hmm), dims=('molecule', 'frame'))
    for molecule in range(len(sort_indices.molecule)):
        mapping = {val: i for i, val in enumerate(sort_indices[molecule].values)}
        mapping.update({val: -1 for val in classification_hmm[molecule, :].values if val < 0})
        classification_hmm_sorted[molecule, :] = np.vectorize(mapping.get)(classification_hmm[molecule, :].values)
    classification_hmm[:] = classification_hmm_sorted[:]

    return state_parameters, transition_matrices, classification_hmm

def trace_classification_model(traces, model):
    classification = np.vstack([model.predict(traces[m].values) for m in traces.molecule.values])
    classification = xr.DataArray(classification, dims=traces.dims)
    return classification

def trace_classification_models(traces, classifications, models):
    new_classifications = []
    for model, xi, ci in zip(models, traces.values, classifications.values):
        if model is not None:
            included_frame_selection = ci >= 0
            xis, _ = split_by_classification(xi, included_frame_selection)
            cis, _ = split_by_classification(ci, included_frame_selection)

            new_classification = []
            for xii, cii in zip(xis, cis):
                if cii[0] >= 0:
                    new_classification.append(model.predict(xii))
                else:
                    new_classification.append(-np.ones_like(cii))
            new_classification = np.hstack(new_classification)
            new_classifications.append(new_classification)
        else:
            new_classifications.append(-np.ones_like(xi))
    return xr.DataArray(np.vstack(new_classifications), dims=('molecule','frame'))

def determine_transition_rates_from_probabilities(number_of_states, transition_probabilities, frame_rate):
    # transition_rates = np.full_like(transition_probabilities, np.nan)
    dims = transition_probabilities.dims
    transition_rates = np.full_like(transition_probabilities, np.nan, dtype=np.complex64)
    transition_probabilities = np.array(transition_probabilities)
    number_of_states = np.array(number_of_states)

    for i in range(len(transition_probabilities)):
        if number_of_states[i] > 0:
            transition_rates[i, :number_of_states[i], :number_of_states[i]] = \
                scipy.linalg.logm(transition_probabilities[i, :number_of_states[i], :number_of_states[i]].T).T

    transition_rates = transition_rates * frame_rate

    return xr.DataArray(transition_rates, dims=dims)

def complex_transition_rates_to_nan(transition_rates):
    is_complex = xr.DataArray((np.iscomplex(transition_rates) & ~np.isnan(transition_rates)).any(axis=2).any(axis=1), dims=('molecule'))
    transition_rates[is_complex, :, :] = np.nan
    return np.real(transition_rates), ~is_complex

def transition_rates_outside_measurement_resolution_to_nan(transition_rates, number_of_frames, frame_rate):
    # For more than two states we likely only have to take the off diagonal components
    off_diagonal_terms = transition_rates.values[:, ~np.eye(*transition_rates.shape[1:], dtype=bool)]
    has_too_low_rate = xr.DataArray((np.abs(off_diagonal_terms) < frame_rate/number_of_frames).any(axis=1), dims='molecule')
    # has_too_low_rate = xr.DataArray(np.diagonal(np.abs(transition_rates), axis1=1, axis2=2).any(axis=1), dims='molecule')
    # has_too_high_rate = (np.abs(ds.transition_rate) > frame_rate).any(axis=2).any(axis=1)
    # transition_rates[has_too_low_rate | has_too_high_rate, :, :] = np.nan
    transition_rates[has_too_low_rate, :, :] = np.nan
    return transition_rates, ~has_too_low_rate #, ~has_too_high_rate


def histogram_1d_state_means(ds, name, save_path, number_of_states=1, state_index=0):
    fig, ax = plt.subplots(figsize=(6.5,3.5), tight_layout=True)
    ds_subset = ds.sel(molecule=ds.number_of_states==number_of_states)
    if state_index > number_of_states-1:
        raise ValueError('State index larger than number of states')

    ax.hist(ds_subset.state_mean.sel(state=state_index), bins=50, range=(0,1))
    ax.set_xlabel('Mean FRET')
    ax.set_ylabel('Molecule count')
    title = name + f' - FRET histogram - state {state_index+1} out of {number_of_states}'
    ax.set_title(title)
    fig.savefig(save_path / (title + '.png'))

    # counts, bins = np.histogram(parameters_one_state[:,0], bins=50, range=(0,1))
    # print("Max at E=", (bins[counts.argmax()]+bins[counts.argmax()+1])/2)

import matplotlib.pyplot as plt
from matplotlib import cm
def histogram_2d_state_means(ds, name, save_path, number_of_states=2, state_indices=[0,1]):
    fig, ax = plt.subplots(figsize=(8,6.5), tight_layout=True)
    ds_subset = ds.sel(molecule=ds.number_of_states==number_of_states)
    for state_index in state_indices:
        if state_index > number_of_states-1:
            raise ValueError('State index larger than number of states')

    ax.hist2d(*ds_subset.state_mean.sel(state=state_indices).T, bins=50, range=((0, 1), (0, 1)))
    ax.set_xlabel(f'Mean FRET - state {state_indices[0]+1}')
    ax.set_ylabel(f'Mean FRET - state {state_indices[1]+1}')
    cax = fig.colorbar(ax.collections[0], ax=ax, label='Molecule count')
    ax.set_aspect(1)
    title = name + f' - FRET histogram - States {state_indices[0]+1} and {state_indices[1]+1} out of {number_of_states}'
    ax.set_title(title)
    fig.savefig(save_path / (title + '.png'))



def histogram_2d_transition_rates(ds, name, save_path, frame_rate, number_of_states=2, state_indices=[0,1]):
    fig, ax = plt.subplots(figsize=(8, 6.5), tight_layout=True)
    ds_subset = ds.sel(molecule=ds.number_of_states == number_of_states)
    for state_index in state_indices:
        if state_index > number_of_states - 1:
            raise ValueError('State index larger than number of states')

    state_A_to_B = ds_subset.transition_rate.sel(from_state=state_indices[0], to_state=state_indices[1])
    state_B_to_A = ds_subset.transition_rate.sel(from_state=state_indices[1], to_state=state_indices[0])

    # ax.hist2d(state_A_to_B, state_B_to_A, bins=50, range=((0, frame_rate), (0, frame_rate)))
    ax.hist2d(state_A_to_B, state_B_to_A, bins=50, range=((0, 16), (0, 16)))
    ax.set_xlabel(f'Transition rate (/s) - state {state_indices[0] + 1}')
    ax.set_ylabel(f'Transition rate (/s) - state {state_indices[1] + 1}')
    cax = fig.colorbar(ax.collections[0], ax=ax, label='Molecule count')
    ax.set_aspect(1)
    title = name + f' - Transition rate histogram - States {state_indices[0] + 1} and {state_indices[1] + 1} out of {number_of_states}'
    ax.set_title(title)
    fig.savefig(save_path / (title + '.png'))


def transition_rate_fit(ds, frame_rate, number_of_states=2, from_state=0, to_state=1):
    # fig, ax = plt.subplots(figsize=(8, 6.5), tight_layout=True)
    ds_subset = ds.sel(molecule=ds.number_of_states == number_of_states)

    transition_rates = ds_subset.transition_rate.sel(from_state=from_state, to_state=to_state).values
    transition_rates = transition_rates[~np.isnan(transition_rates)]

    import scipy.stats
    kernel = scipy.stats.gaussian_kde(transition_rates)
    def gaussian_single(x, a, mean, std):
        # print(x,a,mean,std)
        return a * np.exp(-1/2 * (x-mean)**2 / std**2)

    import scipy.optimize
    x = np.linspace(0,frame_rate,200)
    popt, pcov = scipy.optimize.curve_fit(gaussian_single, x, kernel(x), bounds=((0,-np.inf,0),(np.inf, np.inf, np.inf)))

    a, mean, std = popt
    plt.figure()
    plt.plot(x,kernel(x))
    plt.plot(x,gaussian_single(x, *popt))

    return mean, std




















def decode_traces(model, dataarray):
    classifications = []
    for i in tqdm.tqdm(dataarray.molecule):
        d_i = dataarray.sel(molecule=i).values.T
        log_prob, classification_HMM = model.decode(d_i)
        classifications.append(classification_HMM)
    return xr.DataArray(np.vstack(classifications), dims=('molecule','frame'))


# exp = ta.Experiment(r'D:\20200918 - Test data\Single-molecule data small')
# exp = ta.Experiment(r'P:\SURFdrive\Promotie\Data\Test data')

# ds = xr.open_mfdataset(file_paths, concat_dim='molecule', combine='nested')
# classification_background = ~(ds.intensity < 5000).any('channel')

# file = files_green_laser[0]
# for file in files:
#
#     traces = file.dataset.FRET #ds.intensity.sel(channel=0, drop=True)

# n_components=2, covariance_type="full", n_iter=100

def hmm_traces(traces, initial_boolean_classification=None, **kwargs):
    if initial_boolean_classification is None:
        initial_boolean_classification = [None]*len(traces)

    classification = -xr.ones_like(traces).astype(int)
    classification.name = 'classification'
    hmm_transition_matrix = xr.DataArray(np.ones((len(traces.molecule), 2, 2))*np.nan, coords={},
                                         dims=['molecule', 'from_state', 'to_state'], name='hmm_transition_matrix')
    hmm_means = xr.DataArray(np.ones((len(traces.molecule), 2))*np.nan, coords={}, dims=['molecule','state'], name='hmm_means')
    hmm_log_probs = xr.DataArray(np.ones(len(traces.molecule))*np.nan, coords={}, dims='molecule', name='hmm_log_probabilities')

    for molecule in tqdm(traces.molecule.values):
        hmm_transition_matrix[molecule],  hmm_means[molecule], classification[molecule], hmm_log_probs[molecule] = \
            hmm_trace(traces[molecule], initial_boolean_classification=initial_boolean_classification[molecule], **kwargs)

    return xr.merge([hmm_transition_matrix, hmm_means, hmm_log_probs, classification])


    # hmm_transition_matrix.name = 'hmm_transition_matrix'
    # hmm_transition_matrix.to_netcdf(file.relativeFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')
    # hmm_means.name = 'hmm_means'
    # hmm_means.to_netcdf(file.relativeFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')
    # classification.name = 'classification'
    # classification.to_netcdf(file.relativeFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')

    # nds = xr.Dataset({'hmm_transition_matrix': hmm_transition_matrix, 'hmm_means': hmm_means, 'classification': classification})
    # nds.load()
    # nds.to_netcdf(file.relativeFilePath.with_suffix('.nc'), engine='h5netcdf', mode='a')

# def hmm_traces(traces, **kwargs):
#     hmm_transition_matrix, hmm_means, classification, hmm_log_probs =\
#         xr.apply_ufunc(
#             hmm_trace,
#             traces,
#             kwargs=kwargs,
#             input_core_dims=[['frame']],
#             output_core_dims=[['from_state','to_state'],['state'],['frame'],[]],
#             # dask="parallelized",
#             output_dtypes=[np.ndarray, np.ndarray, np.ndarray, float],
#             dask_gufunc_kwargs=dict(output_sizes={"from_state": 2, "to_state": 2, 'state': 2}),
#             vectorize=True
#         )
#     hmm_transition_matrix.name = 'hmm_transition_matrix'
#     hmm_means.name = 'hmm_means'
#     classification.name = 'classification'
#     hmm_log_probs.name = 'hmm_log_probabilities'
#     return xr.merge([hmm_transition_matrix, hmm_means, hmm_log_probs, classification])


def hmm_trace(trace, initial_boolean_classification=None, **kwargs):
    # TODO: Test handling of initial_boolean_classification
    # print(trace)
    # FRET = np.atleast_2d(traces.sel(molecule=molecule_index)).T
    trace_ = np.array(trace).reshape(-1, 1)
    # print(trace_)
    if initial_boolean_classification is not None:
        # initial_boolean_classification = np.ones_like(trace).astype(bool)
        trace_ = trace_[initial_boolean_classification]
        segment_lengths = [sum(1 for _ in group) for value, group in groupby(initial_boolean_classification) if value]
    else:
        segment_lengths = None
    model = hmm.GaussianHMM(**kwargs)
    model.fit(trace_, segment_lengths)
    try:
        log_prob, classification_HMM = model.decode(trace_)
        #classification_HMM = xr.DataArray(classification_HMM, coords=trace.coords)
        if initial_boolean_classification is not None:
            classification_HMM[~initial_boolean_classification] = -1
    except ValueError:
        print('!!!')
        log_prob = np.nan
        classification_HMM = -np.ones_like(trace).astype(int)

    return model.transmat_, model.means_.squeeze(), classification_HMM, log_prob



# p_stay = np.diagonal(hmm_transition_matrix, axis1=1, axis2=2)
#
# dt = 0.1 #s
# tau = -dt/np.log(p_stay)


# for file_path in file_paths:
#     with xr.open_dataset(file_path) as ds:
#         # ds = ds.reset_index('molecule', drop=True)
#         # # ds = ds.reset_index('dim_1', drop=True)
#         # ds = ds.reset_index('frame', drop=True)
#         # ds = ds.reset_index('from_state', drop=True)
#         # ds = ds.reset_index('to_state', drop=True)
#         ds.load()
#     ds['channel'] = [0,1]
#
#     ds.to_netcdf(file_path, engine='h5netcdf', mode='w')





from objectlist import ObjectList

from papylio.file import File
from papylio.netcdf_operations import merge_datasets, reorder_datasets_using_sequence_subset

import numpy as np
import xarray as xr
import matplotlib.pyplot as plt

# TODO: Make sure that the collection can only contain File objects
class FileCollection(ObjectList):
    def __getattr__(self, item):
        # TODO: Replace this method by map?
        attrs = super(FileCollection, self).__getattr__(item)
        if callable(attrs):  # Test if attrs is a function, if this is the case add another function to concatenate possible xarray objects.
            def f2(*args, **kwargs):
                output = attrs(*args, **kwargs)
                if output is not None and isinstance(output[0], xr.DataArray):
                    output = xr.concat(output, dim=output[0].dims[0])
                return output
            return f2

        elif isinstance(attrs[0], xr.DataArray):
            attrs = xr.concat(attrs, dim=attrs[0].dims[0])
        elif isinstance(attrs[0], xr.Dataset):
            if 'molecule' in attrs[0].dims:
                attrs = xr.concat(attrs, dim='molecule')
            elif item == 'dwells':
                #TODO is this still necessary or is this done by the xr.DataArray elif?
                attrs = xr.concat(attrs, dim='dwell')
        return attrs

    @property
    def experiment(self):
        return self[0].experiment

    def select(self, search_string, variable='relativeFilePath'):
        # TODO: Make this accept keyword arguments where the key is the variable and the item is the search_string.
        # TODO: Make this accept mulitple keyword arguments.
        return self[getattr(self, variable).str.regex(search_string)]

    def show_histogram(self, *args, **kwargs):
        figure, axis = File.show_histogram(self.serial, *args, **kwargs)
        axis.set_title('')
        return figure, axis

    def histogram_FRET_intensity_total(self, **kwargs):
        File.histogram_FRET_intensity_total(self, **kwargs)

    @property
    def cycle_time(self):
        return self[0].cycle_time

    def analyze_dwells_combined(self, *args, name='', **kwargs):
        return File.analyze_dwells(self.serial, *args, name=name, **kwargs)

    def print(self):
        for i, file in enumerate(self):
            print(f"{i:3d}.  {file.relativeFilePath}")

    def merge_datasets(self, filepath_out=None, init_file_index=0, with_selected_only=False, with_sequence_only=False):
        #TODO: remove sequencing part, or move to the sequencing plugin
        if filepath_out is None:
            filepath_out = self[0].absoluteFilePath.parent / 'merged_dataset.nc'
        filepaths_in = self.serial.absoluteFilePath.with_suffix('.nc')
        merge_datasets(filepaths_in, filepath_out, concat_dim='molecule', init_file=filepaths_in[init_file_index],
                       with_selected_only=with_selected_only, with_sequence_only=with_sequence_only)

    def reorder_datasets_using_sequence_subset(self, folderpath_out):
        files = self[self.has_sequencing_match]
        filepaths_in = files.serial.absoluteFilePath.with_suffix('.nc')
        reorder_datasets_using_sequence_subset(filepaths_in, folderpath_out, concat_dim='molecule')


import re
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

def plot_sequence_density(sequences, expected_seq=None, start=None, end=None, is_circular=False,
                          row_length=None, figure=None, save=False, title=''):
    sequences = sequences.view('U1').reshape(-1, len(sequences[0]))
    if is_circular:
        sequences = np.hstack([sequences, sequences[:,0:1]])
    sequence = sequences[~np.any(sequences == 'N', axis=1)]
    # sequence_int = sequence.view('uint8')
    # sequence_df = pd.DataFrame(sequence_int).iloc[:, slice(start, end)]
    sequence_df = pd.DataFrame(sequence)
    if expected_seq is not None:
        expected_seq = expected_seq[slice(start, end)]

    total_length = sequence_df.shape[1]

    if not row_length:
        row_length = total_length

    n_rows = total_length // row_length
    if figure is None:
        figure = plt.figure(figsize=(np.max([row_length / 4, 6]), 3 * n_rows), layout='constrained')

    figure.subplots(n_rows, 1, sharey=False)
    axes = figure.axes

    bases = ['A', 'T', 'C', 'G']
    index = [(b1, b2) for b1 in bases for b2 in bases]
    # out = pd.DataFrame(index=index)
    for r in np.arange(n_rows):
        for i in np.arange((r * row_length), ((r + 1) * row_length) - 1):
            # print(i)
            # out[i] = (test.iloc[:,i:(i+2)].value_counts(normalize=True))
            base_transition_count = (sequence_df.iloc[:, i:(i + 2)].value_counts(normalize=True)).reindex(index)
            ys = base_transition_count.index.to_list()
            zs = base_transition_count.values
            x = (i, i + 1)
            for y, z in zip(ys, zs):
                axes[r].plot(x, y, lw=z * 10, c='b', solid_capstyle='round')
            # ax.plot((i, i + 1), [0, 0])
        # expected_seq=sequences_per_read['Read1']['HJ1']
        if expected_seq is not None:
            # axis.set_xticklabels(list(expected_seq))
            # axis.set_xticks(np.arange(len(expected_seq)))
            xs = np.where(np.array(list(expected_seq)) != 'N')[0]
            x0 = 0
            for s in re.split('N', expected_seq):
                if s:
                    x = xs[x0:(x0 + len(s))]
                    y = np.array(list(s), dtype='S1')  # .view('uint8')
                    axes[r].plot(x, y, c='r')
                    x0 += len(s)
        axes[r].set_xlim(r * row_length, (r + 1) * row_length - 1)
        axes[r].set_ylabel('Base')
        # axes[r].xaxis.set_minor_locator(plt.MultipleLocator(1))

    axes[0].set_title(title)
    axes[-1].set_xlabel('Position')
    if is_circular:
        xticklabels = np.arange(sequences.shape[1])
        xticklabels[-1] = 0
        axes[-1].set_xticks(np.arange(len(xticklabels)))
        axes[-1].set_xticklabels(xticklabels)

    # for axis in axes:
    #     bases = list('ATGC')
    #     axis.set_yticks(np.array(bases, dtype='S1').view('uint8'))
    #     axis.set_yticklabels(bases)

    # figure.tight_layout(pad=1)
    if save:
        title = title.replace('>', 'gt').replace('<', 'st')
        figure.savefig(f'{title}_seqdensity.png')


import numpy as np
from pathlib import Path
import h5py
import re
import pandas as pd
import netCDF4
import tabulate
# import h5netcdf.legacyapi as netCDF4
# import h5netcdf
import tqdm
from contextlib import ExitStack
import xarray as xr
import re
import matplotlib.pyplot as plt

from papylio.plugins.sequencing.plotting import plot_cluster_locations_per_tile
#
# class DatasetPointer:
#     class_attributes = ['file_path', '_dataset']
#
#     def __init__(self, file_path):
#         self.__dict__['file_path'] = file_path
#         self.__dict__['_dataset'] = None
#
#     def __enter__(self):
#         self._dataset = xr.open_dataset(self.file_path, engine='netcdf4')
#         return self._dataset
#
#     def __exit__(self, exc_type, exc_val, exc_tb):
#         self._dataset.close()
#         self._dataset = None
#
#     def __repr__(self):
#         with self as ds:
#             return str(ds)
#
#     def __getattr__(self, item):
#         print(item)
#         with self as ds:
#             return ds[item].load()
#
#     def __setattr__(self, key, value):
#         if key in DatasetPointer.class_attributes:
#             super().__setattr__(key, value)
#         else:
#             print('__setattr__', key, value)
#             value = xr.DataArray(value)
#             value.name = key
#             value.to_netcdf(self.file_path, mode='a', engine='h5netcdf')
# For testing
# import pytest
# from trace_analysis.plugins.sequencing.sequencing_data import DatasetPointer
#
#
# def test_DatasetPointer(shared_datadir):
#     ds = DatasetPointer(shared_datadir / 'BN_TIRF_sequencing/HJ_scan TIRF 561 0300_sequencing_data.nc')
#     ds.tile



# Update class with new sam analysis function below
class SequencingData:

    reagent_kit_info = {'v2':       {'number_of_tiles': 14, 'number_of_surfaces': 2},
                        'v2_micro': {'number_of_tiles':  4, 'number_of_surfaces': 2},
                        'v2_nano':  {'number_of_tiles':  2, 'number_of_surfaces': 1},
                        'v3':       {'number_of_tiles': 19, 'number_of_surfaces': 2}}

    # @classmethod
    # def load(cls):
    #     cls()

    def __init__(self, file_path=None, dataset=None, name='', reagent_kit='v3', load=True, file_kwargs={}, save_path=None):
        if file_path is not None:
            file_path = Path(file_path)
            if file_path.suffix == '.nc':
                # with xr.open_dataset(file_path.with_suffix('.nc'), engine='h5netcdf') as dataset:
                #     self.dataset = dataset.load().set_index({'sequence': ('tile','x','y')})
                if load:
                    self.dataset = xr.load_dataset(file_path.with_suffix('.nc'), engine='netcdf4')
                else:
                    self.dataset = xr.open_dataset(file_path.with_suffix('.nc'), engine='netcdf4')#, chunks=10000)
                    # self.dataset = DatasetPointer(file_path.with_suffix('.nc'))
                # self.dataset = self.dataset.set_index({'sequence': ('tile', 'x', 'y')})
                # self.dataset.coordsupdate
                # self.dataset.update(
                #     {'sequence': pd.MultiIndex.from_frame(self.dataset[['tile', 'x', 'y']].to_pandas())})

            else:
                data = pd.read_csv(file_path, delimiter='\t')
                data.columns = data.columns.str.lower()
                data = data.set_index(['tile', 'x', 'y'])
                data = data.drop_duplicates() # TODO: It would be better to do this when converting the sam file.
                data.index.name = 'sequence' # Do this after drop_duplicates!
                self.dataset = xr.Dataset(data) #.reset_index('sequence', drop=True)

                if name == '':
                    name = Path(file_path).name

        elif dataset is not None:
            self.dataset = dataset
        else:
            raise ValueError('Either file_path or data should be given')

        self.name = name
        if save_path is None:
            self.save_path = file_path.parent
        else:
            self.save_path = save_path

        self.reagent_kit = reagent_kit
        self.reagent_kit_info = SequencingData.reagent_kit_info[reagent_kit]

        self._tiles = None


    def __getattr__(self, item):
        #try:
        # print('seqdata=' + item)
        if 'dataset' in self.__dict__.keys() and hasattr(self.dataset, item):
            return getattr(self.dataset, item)
        #except AttributeError:
        #    super().__getattribute__(item)
        else:
            raise AttributeError
            # super(SequencingData, self).__getattribute__(item)

        # try:
        #     return getattr(self.dataset, item)
        # except AttributeError:
        #     super().__getattribute__(item)

    def __getitem__(self, item):
        return SequencingData(dataset=self.dataset[dict(sequence=item)], reagent_kit=self.reagent_kit, save_path=self.save_path)

    def __repr__(self):
        return (f'{self.__class__.__name__}({self.name})')

    def __len__(self):
        return len(self.dataset.sequence)

    @property
    def coordinates(self):
        return xr.DataArray([self.dataset['x'], self.dataset['y']], dims=('dimension','sequence'),
                            coords={'dimension': ['x','y'], 'sequence': self.dataset.sequence, 'tile': self.dataset.tile}, name='coordinates').T
        # return self.dataset[['x','y']].to_array(dim='dimension', name='coordinates').transpose('sequence',...)
            # xr.DataArray(self.data[['x','y']], dims=['sequence', 'dimension'], name='coordinates')\
            # .reset_index('sequence', drop=True)

    @property
    def tile_numbers(self):
        return np.unique(self.dataset.tile)

    # @property
    # def tiles(self):
    #     if not self._tiles:
    #         # Perhaps more elegant if this returns SequencingData objects [25-10-2021 IS]
    #         self._tiles = [Tile(tile, tile_coordinates) for tile, tile_coordinates in self.coordinates.groupby('tile')]
    #     return self._tiles

    def sel(self, *args, **kwargs):
        return SequencingData(dataset=self.dataset.sel(*args, **kwargs))

    # def classify_sequences(self, variable=''):

    def reference_distribution(self, save=True, report=True):
        da_reference_name = self.dataset.reference_name.load()
        da_sequence_subset = self.dataset.sequence_subset.load()

        reference_names, counts = np.unique(da_reference_name, return_counts=True)

        analysis_reference = xr.Dataset(coords={'reference_name': reference_names})
        analysis_reference['reference_count'] = xr.DataArray(counts, dims='reference_name')
        analysis_reference['reference_fraction'] = xr.DataArray(counts / counts.sum(), dims='reference_name')

        full_subset_counts = np.zeros(len(reference_names), dtype='int64')
        for i, reference_name in enumerate(reference_names):
            da_sequence_subset_sel = da_sequence_subset.sel(sequence=da_reference_name == reference_name)
            full_subset_counts[i] = (~da_sequence_subset_sel.str.contains('-')).sum().item()
        analysis_reference['full_subset_count'] = xr.DataArray(full_subset_counts, dims='reference_name')
        analysis_reference['full_subset_fraction'] = analysis_reference['full_subset_count'] / analysis_reference[
            'reference_count']

        # reference_names[reference_names=='*'] = 'Unmapped'
        # analysis_reference.reindex({'reference': reference_names})
        if save:
            analysis_reference.to_netcdf(self.save_path / 'reference_distribution.nc')

        if report:
            string = 'Mapped sequences \n================\n\n' + \
                     tabulate.tabulate(analysis_reference.to_pandas(),
                                       headers=["Reference name", "Reference\nCount", "\nPercentage",
                                                "Full subset\nCount", "\nPercentage"],
                                       floatfmt=(None, ".0f", ".2%", ".0f", ".2%"))
            print(string)
            self.add_to_report_file(string)

        return analysis_reference

    def show_reference_distribution(self, save=True):
        analysis_reference = self.reference_distribution(save=save)

        fig, ax = plt.subplots(layout='tight', figsize=(analysis_reference.reference_name.size,3))
        ax.bar(analysis_reference['reference_name'], analysis_reference['reference_count'], fc='grey')
        total = analysis_reference['reference_count'].sum().item()
        secax = ax.secondary_yaxis('right', functions=(lambda x: x/total, lambda x: x*total))
        ax.ticklabel_format(axis='y', style='sci', scilimits=(-3,3))
        ax.set_xlabel('Reference')
        ax.set_ylabel('Count')
        secax.set_ylabel('Fraction')
        ax.set_title('Reference distribution')

        if hasattr(analysis_reference, 'full_subset_count'):
            ax.bar(analysis_reference['reference_name'], analysis_reference['full_subset_count'], fc='grey', ec='white', lw=0, label='Complete subset', hatch='/////')
            ax.set_ylim(0, ax.get_ylim()[1]*1.2)
            ax.legend(frameon=False, loc='upper right') # bbox_to_anchor=(1, 1),

        if save:
            fig.savefig(self.save_path / 'reference_distribution.png')
            fig.savefig(self.save_path / 'reference_distribution.pdf')

    def distribution(self, variable='reference_name', save=True, report=True, remove_empty_strings=False):
        da_variable = self.dataset[variable].load()
        da_sequence_subset = self.dataset.sequence_subset.load()

        if remove_empty_strings:
            selection = da_variable != ''
            da_variable = da_variable[selection]
            da_sequence_subset = da_sequence_subset[selection]

        names, counts = np.unique(da_variable, return_counts=True)

        ds = xr.Dataset(coords={variable: names})
        ds['count'] = xr.DataArray(counts, dims=variable)
        ds['fraction'] = xr.DataArray(counts / counts.sum(), dims=variable)

        full_subset_counts = np.zeros(len(names), dtype='int64')
        for i, name in enumerate(names):
            da_sequence_subset_sel = da_sequence_subset.sel(sequence=da_variable == name)
            full_subset_counts[i] = (~da_sequence_subset_sel.str.contains('-')).sum().item()
        ds['full_subset_count'] = xr.DataArray(full_subset_counts, dims=variable)
        ds['full_subset_fraction'] = ds['full_subset_count'] / ds['count']

        # reference_names[reference_names=='*'] = 'Unmapped'
        # analysis_reference.reindex({'reference': reference_names})
        if save:
            ds.to_netcdf(self.save_path / f'{variable}_distribution.nc')

        if report:
            string = 'Mapped sequences \n================\n\n' + \
                     tabulate.tabulate(ds.to_pandas(),
                                       headers=[variable, "Count", "\nPercentage",
                                                "Full subset\nCount", "\nPercentage"],
                                       floatfmt=(None, ".0f", ".2%", ".0f", ".2%"))
            print(string)
            self.add_to_report_file(string)

        return ds

    def show_distribution(self, variable='reference_name', save=True):
        ds = self.distribution(variable=variable, save=save)

        #TODO: Move to plotting.py ?
        fig, ax = plt.subplots(layout='tight', figsize=(ds[variable].size+1, 3))
        ds[variable] = ds[variable].astype('U').str.replace('_','\n')
        ax.bar(ds[variable], ds['count'], fc='grey')
        total = ds['count'].sum().item()
        secax = ax.secondary_yaxis('right', functions=(lambda x: x / total, lambda x: x * total))
        ax.ticklabel_format(axis='y', style='sci', scilimits=(-3, 3))
        ax.set_xlabel(variable)
        ax.set_ylabel('Count')
        secax.set_ylabel('Fraction')
        ax.set_title(f'{variable} distribution')

        if hasattr(ds, 'full_subset_count'):
            ax.bar(ds[variable], ds['full_subset_count'], fc='grey', ec='white',
                   lw=0, label='Complete subset', hatch='/////')
            ax.set_ylim(0, ax.get_ylim()[1] * 1.2)
            ax.legend(frameon=False, loc='upper right')  # bbox_to_anchor=(1, 1),

        if save:
            fig.savefig(self.save_path / f'{variable}_distribution.png')
            fig.savefig(self.save_path / f'{variable}_distribution.pdf')

    def get_sequences(self, reference_name=None, variable='read1_sequence', remove_incomplete_sequences=True):
        dataset = self.dataset
        if reference_name is not None:
            dataset = dataset.sel(sequence=self.dataset.reference_name == reference_name)

        sequences = dataset[variable].load()

        if remove_incomplete_sequences:
            is_complete = ~np.array(['-' in sequence for sequence in sequences.values])
            sequences = sequences[is_complete]

        return sequences

    def sequence_count(self, variable='sequence_subset',  remove_incomplete_sequences=True, save=True):
        sequences = self.get_sequences(variable=variable, remove_incomplete_sequences=remove_incomplete_sequences)
        n, c = np.unique(sequences, return_counts=True)
        sequence_count = xr.DataArray(c, coords={variable: n})

        sequence_count.name = f'{self.name} - sequence_count - {variable}'
        sequence_count.attrs['variable'] = variable
        sequence_count.attrs['remove_incomplete_sequences'] = str(remove_incomplete_sequences)

        if save:
            sequence_count.to_netcdf(self.save_path / f'{self.name} - sequence_count - {variable}.nc')

        return sequence_count

    def base_composition(self, variable='read1_sequence', positions=None, remove_incomplete_sequences=True, save=True):
        sequences = self.get_sequences(variable=variable, remove_incomplete_sequences=remove_incomplete_sequences)
        if positions is None:
            positions = np.arange(len(sequences[0].item()))

        base_count = xr.DataArray(0, dims=['position', 'base'],
                                  coords={'position': positions, 'base': ['A', 'T', 'C', 'G']})

        for p in tqdm.tqdm(base_count.position.values):
            base_count[dict(position=p)] = \
                [(sequences.str.get(p) == b).sum().item() for b in base_count.base.values]

        base_fractions = base_count / len(sequences)

        base_composition = xr.Dataset()
        base_composition['base_count'] = base_count
        base_composition['base_fraction'] = base_fractions

        base_composition.attrs['variable'] = variable
        base_composition.attrs['remove_incomplete_sequences'] = str(remove_incomplete_sequences)

        if save:
            base_composition.to_netcdf(self.save_path / f'{self.name} - base_composition - {variable}.nc')

        return base_composition

    def show_base_composition(self, save=True, **kwargs):
        base_composition = self.base_composition(**kwargs)
        import logomaker

        figure, axis = plt.subplots(figsize=(min(1+base_composition.position.size/2, 10), 2.5), layout='constrained')
        nn_logo = logomaker.Logo(base_composition.base_fraction.to_pandas(), stack_order='fixed', ax=axis)
        axis.set_ylabel('Fraction')
        axis.set_xlabel('Position')
        variable = base_composition.attrs['variable']
        axis.set_title(f'{self.name} - base composition - {variable}')
        if save:
            figure.savefig(self.save_path / f'{self.name} - base_composition - {variable}.png')
            figure.savefig(self.save_path / f'{self.name} - base_composition - {variable}.pdf')

    def basepair_count(self, variable='read1_sequence', positions=None, remove_incomplete_sequences=True, save=True):
        filepath = self.save_path / f'{self.name} - basepair_count - {variable}.nc'
        if filepath.exists():
            basepair_count = xr.load_dataarray(filepath, engine='netcdf4')
            return basepair_count

        sequences = self.get_sequences(variable=variable, remove_incomplete_sequences=remove_incomplete_sequences)
        if positions is None:
            positions = np.arange(len(sequences[0].item()))

        bases = ['A', 'T', 'C', 'G']
        basepair_count = xr.DataArray(0, dims=['position_0', 'position_1', 'base_0', 'base_1'],
                                      coords={'position_0': positions, 'position_1': positions, 'base_0': bases,
                                              'base_1': bases})
        basepair_count.name = f'{self.name} - basepair_count - {variable}'
        basepair_count.attrs['variable'] = variable
        basepair_count.attrs['remove_incomplete_sequences'] = str(remove_incomplete_sequences)


        for position_0 in tqdm.tqdm(positions):
            bases_position_0 = sequences.str.get(position_0)
            for position_1 in positions:
                bases_position_1 = sequences.str.get(position_1)
                for base_0 in bases:
                    for base_1 in bases:
                        basepair_count.loc[
                            dict(position_0=position_0, position_1=position_1, base_0=base_0, base_1=base_1)] = \
                            ((bases_position_0 == base_0) & (bases_position_1 == base_1)).sum()
        if save:
            basepair_count.to_netcdf(self.save_path / (basepair_count.name + '.nc'))

        return basepair_count

    def new_report_file(self):
        report = open(self.save_path / 'report.txt', 'w')
        report.close()

    def add_to_report_file(self, string):
        report = open(self.save_path / 'report.txt', 'a')
        report.write(string)
        report.close()

    def plot_cluster_locations_per_tile(self, save_filepath=None):
        # TODO: Fix bug self.dataset[['x','y']]
        plot_cluster_locations_per_tile(self.dataset[['tile','x','y']], **self.reagent_kit_info, save_filepath=save_filepath)

    def save(self, filepath):
        self.dataset.reset_index('sequence').to_netcdf(filepath, engine='netcdf4', mode='w')

class Tile:
    def __init__(self, number, coordinates):
        self.name = str(number)
        self.number = number
        self.coordinates = coordinates

    def __repr__(self):
        return (f'{self.__class__.__name__}({self.name})')



def read_sam(sam_filepath, add_aligned_sequence=False, extract_sequence_subset=False, read_name='read1'):
    sam_filepath = Path(sam_filepath)

    # Check number of header lines:
    with Path(sam_filepath).open('r') as file:
        number_of_header_lines = 0
        while file.readline()[0] == '@':
            number_of_header_lines += 1

    df = pd.read_csv(sam_filepath,
                     delimiter='\t', skiprows=number_of_header_lines, usecols=range(11), header=None,
                     names=['sequence_identifier', 'sam_flag', 'reference_name', 'position', 'mapping_quality', 'cigar_string',
                            'mate_reference_name', 'mate_position', 'template_length', name+'_sequence', name+'_quality'])
    df = df.drop_duplicates(subset='sequence_identifier', keep='first', ignore_index=False)
    df.index.name = 'sequence'
    df_split = df['sequence_identifier'].str.split(':', expand=True)
    df_split.columns = ['instrument', 'run', 'flowcell', 'lane', 'tile', 'x', 'y']
    df_split = df_split.astype({'run': int, 'lane': int, 'tile': int, 'x': int, 'y': int})
    df_joined = df_split.join(df.iloc[:, 1:])

    if add_aligned_sequence:
        df_joined[[read_name+'_sequence_aligned', read_name+'_quality_aligned']] = \
            df_joined.apply(get_aligned_sequence_from_row, axis=1, result_type='expand', name=name)

    if extract_sequence_subset:
        df_joined['sequence_subset'] = extract_positions(df_joined[read_name+'_sequence_aligned'],
                                                         extract_sequence_subset)
        df_joined['quality_subset'] = extract_positions(df_joined[read_name+'_quality_aligned'],
                                                        extract_sequence_subset)

    return df_joined


def read_sam_header(sam_filepath):
    header_dict = {'HD': [], 'SQ': [], 'RG': [], 'PG':[], 'CO':[]}
    with Path(sam_filepath).open('r') as file:
        number_of_header_lines = 0
        # read_line = file.readline().split('\t')
        while (read_line := file.readline().rstrip('\n').split('\t'))[0][0] == '@':
        # while read_line[0][0] == '@':
            line_data = {item.split(':')[0]: item.split(':')[1] for item in read_line[1:]}
            header_dict[read_line[0][1:]].append(line_data)
            # read_line = file.readline().split('\t')

            number_of_header_lines += 1

    return number_of_header_lines, header_dict
#
# def number_of_lines(filepath):
#     def _make_gen(reader):
#         while True:
#             b = reader(2 ** 16)
#             if not b: break
#             yield b
#
#     with open(filepath, "rb") as f:
#         count = sum(buf.count(b"\n") for buf in tqdm.tqdm(_make_gen(f.raw.read)))
#     return count

def number_of_lines(filepath):
    return sum(1 for line in tqdm.tqdm(filepath.open('r')))

def number_of_sequence_alignments(sam_filepath):
    return sum(1 if line[0] != '@' else 0 for line in tqdm.tqdm(sam_filepath.open('r')))

def bitwise_flag(flag, bit):
    flag = np.array(flag)
    return (flag & (2**bit)).astype(bool)

def number_of_primary_sequence_alignments(sam_filepath):
    sam_filepath = Path(sam_filepath)
    compiled_regex = re.compile('(?<=\t)\d*(?=\t)')
    is_primary_alignment_list = []
    for line in tqdm.tqdm(sam_filepath.open('r'), 'Determine number of primary alignments'):
        if line[0] != '@': # Remove header lines
            is_primary_alignment = not bitwise_flag(int(compiled_regex.search(line).group()), 11)
            is_primary_alignment_list.append(is_primary_alignment)
    return np.sum(is_primary_alignment_list)

    # return sum(bitwise_flag(int(compiled_regex.search(line).group()), 11)
    #            if line[0]!='@' else 0 for line in tqdm.tqdm(sam_filepath.open('r')))


def parse_sam(sam_filepath, read_name='read1', remove_duplicates=True, add_aligned_sequence=False, extract_sequence_subset=False,
              chunksize=10000, write_csv=False, write_nc=True, write_filepath=None):
    sam_filepath = Path(sam_filepath)

    # Check number of header lines:
    number_of_header_lines, header_dict = read_sam_header(sam_filepath)

    name_dict = {'*': 0, '=': 1}
    name_dict.update({SQ_dict['SN']:i+2 for i, SQ_dict in enumerate(header_dict['SQ'])})

    max_reference_length = np.array([int(d['LN']) for d in header_dict['SQ']]).max()

    number_of_sequences = number_of_primary_sequence_alignments(sam_filepath)

    end_index = 0

    with ExitStack() as stack:

        with pd.read_csv(sam_filepath,
                         delimiter='\t', skiprows=number_of_header_lines, usecols=range(11), header=None,
                         names=['sequence_identifier', 'sam_flag', 'reference_name', 'position', 'mapping_quality',
                                'cigar_string', 'mate_reference_name', 'mate_position', 'template_length',
                                read_name + '_sequence', read_name + '_quality'],
                         chunksize=chunksize) as reader:

            for i, chunk in tqdm.tqdm(enumerate(reader), 'Parse sam file', total=number_of_sequences//chunksize+1):
                df_chunk = chunk
                df_chunk.index.name = 'sequence'
                if remove_duplicates:
                    df_chunk = df_chunk[~bitwise_flag(df_chunk.sam_flag, 11)]
                    # df_chunk = df_chunk[df_chunk.sam_flag//2048 == 0]
                df_split = df_chunk['sequence_identifier'].str.split(':', expand=True)
                df_split.columns = ['instrument', 'run', 'flowcell', 'lane', 'tile', 'x', 'y']
                df_split = df_split.astype({'run': int, 'lane': int, 'tile': int, 'x': int, 'y': int})
                df = df_split.join(df_chunk.iloc[:, 1:])

                if add_aligned_sequence:
                    df[[read_name+'_sequence_aligned', read_name+'_quality_aligned']] = \
                        df.apply(get_aligned_sequence_from_row, axis=1, result_type='expand', read_name=read_name,
                                 reference_range=(0, max_reference_length))

                if extract_sequence_subset:
                    df['sequence_subset'] = extract_positions(df[read_name+'_sequence_aligned'], extract_sequence_subset)
                    df['quality_subset'] = extract_positions(df[read_name+'_quality_aligned'], extract_sequence_subset)

                if write_filepath is None:
                    write_filepath = sam_filepath.with_suffix('')

                if write_csv:
                    if i == 0:
                        csv_filepath = write_filepath.with_suffix('.csv')
                        df.to_csv(csv_filepath, header=True, mode='w')
                    else:
                        df.to_csv(csv_filepath, header=False, mode='a')

                if write_nc:
                    if i == 0:
                        nc_filepath = write_filepath.with_suffix('.nc')
                        # with netCDF4.Dataset(nc_filepath, 'w'):
                        #     nc_file.createDimension('sequence', None)

                        nc_file = stack.enter_context(netCDF4.Dataset(nc_filepath, 'w'))
                        nc_file.createDimension('sequence', number_of_sequences)
                        for name, datatype in df.dtypes.items():
                            if name in ['instrument','run','flowcell']:
                                setattr(nc_file, name, df[name][0])
                            elif name in ['reference_name', 'mate_reference_name']:
                                # nc_file.createVariable(name, np.uint8, ('sequence',))
                                # nc_file[name].enum_dict = name_dict
                                size = np.array(list(name_dict.keys())).astype('S').itemsize
                                create_string_variable_in_nc_file(nc_file, name, dimensions=('sequence',), size=size)
                            elif (read_name+'_sequence') in name or (read_name+'_quality') in name or 'subset' in name:
                                size = len(df[name][0])
                                create_string_variable_in_nc_file(nc_file, name, dimensions=('sequence',), size=size)
                            elif name in ['cigar_string']:
                            # if datatype == np.dtype('O'):
                                # nc_file.createDimension(name + '_size', None)
                                # nc_file.createVariable(name, 'S1', ('sequence', name + '_size'), chunksizes=(10000, 1))
                                # nc_file[name]._Encoding = 'utf-8'
                                create_string_variable_in_nc_file(nc_file, name, dimensions=('sequence',), chunksizes=(1, 25))
                            else:
                                nc_file.createVariable(name, datatype, ('sequence', ))

                    start_index = end_index
                    end_index = start_index + len(df)
                    for name, datatype in df.dtypes.items():
                        # print(name)
                        if name in ['instrument', 'run', 'flowcell']:
                            continue
                        elif datatype == np.dtype('O'):
                            size = np.max([2, nc_file.dimensions[name+'_size'].size, df[name].str.len().max()])
                            nc_file[name][start_index:end_index] = df[name].values.astype(f'S{size}')
                            # nc_file[name][old_size:, :] = df[name].values.astype(f'S{size}').view('S1').reshape(-1, size)
                        else:
                            nc_file[name][start_index:end_index] = df[name].values

                    # To set the coords for xarray.
                    nc_file.setncattr('coordinates', 'tile x y')

            # Use this if xarray should open the file with standard datatype "|S" instead of "object"
            # for name, datatype in df.dtypes.items():
            #     print(name)
            #     if datatype == np.dtype('O'):
            #         delattr(nc_file[name], '_Encoding')


                    #
                    #     nc_file['cigar_string'][:] = df[0:200].cigar_string.astype('S').values
                    #     # delattr(nc_file['cigar_string'], '_Encoding')
                    #     nc_file.close()
                    #
                    #     test = xr.load_dataset(write_filepath.with_suffix('.nc'))
                    #
                    #
                    #     nc_file = stack.enter_context(h5netcdf.File(nc_filepath, 'a'))
                    #     nc_file.dimensions = {'sequence': None}
                    #     for name, datatype in df.dtypes.items():
                    #         if datatype == np.dtype('O'):
                    #             # datatype = h5py.string_dtype(encoding='utf-8')
                    #             datatype = 'S10'
                    #         else:
                    #             nc_file.create_variable(name, ('sequence',), data=None, dtype=datatype, chunks=(chunksize,))
                    # old_size = nc_file.dimensions['sequence'].size
                    # new_size = old_size + len(df)
                    # nc_file.resize_dimension('sequence', new_size)
                    # added_data_slice = slice(old_size, new_size)
                    # for name in df.columns:
                    #     nc_file[name][added_data_slice] = df[name].values

    #
    # encoding = {key: {'dtype': '|S'} for key in ds.keys() if ds[key].dtype == 'object'}
    #
    # keys = ['instrument', 'run', 'flowcell', 'lane', 'tile', 'x', 'y', 'sam_flag', 'contig_name', 'first_base_position', 'mapping_quality', 'cigar_string', 'mate_name', 'mate_position', 'template_length', 'read_sequence', 'read_quality', 'read_sequence_aligned', 'read_quality_aligned', 'sequence_subset', 'quality_subset', 'index1_sequence', 'index1_quality']
    # with xr.open_dataset(
    #         r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220607 - Sequencer (MiSeq)\Analysis\sequencing_data.nc') as ds:
    #     keys = list(ds.keys())
    # for key in keys:
    #     print(key)
    #     with xr.open_dataset(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220607 - Sequencer (MiSeq)\Analysis\sequencing_data.nc') as ds:
    #         da = ds[key].load()
    #         if da.dtype == 'object':
    #             encoding = {key: {'dtype': '|S'}}
    #         else:
    #             encoding = {}
    #         da.to_netcdf(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20220607 - Sequencer (MiSeq)\Analysis\sequencing_data_S.nc', engine='h5netcdf', mode='a', encoding=encoding)


# for key in ds.keys():
#     if ds[key].dtype == 'object':
#         print(key)


def fastq_data(fastq_filepath, read_name):
    tile_list = []
    x_list = []
    y_list = []
    sequence = []
    quality = []
    expr = re.compile('[:@ \n]')
    with Path(fastq_filepath).open('r') as fq_file:
        for line_index, line in enumerate(tqdm.tqdm(fq_file)):
            if line_index % 4 == 0:
                name = line.strip()
                instrument, run, flowcell, lane, tile, x, y = expr.split(name)[1:8]
                # sequence_index = numbered_index.loc[dict(sequence=(int(tile), int(x), int(y)))].item()
                tile_list.append(int(tile))
                x_list.append(int(x))
                y_list.append(int(y))

            if line_index % 4 == 1:
                sequence.append(line.strip())
            if line_index % 4 == 3:
                quality.append(line.strip())
    ds = xr.Dataset({read_name+'_sequence': ('sequence', sequence), read_name+'_quality': ('sequence', quality)},
                      coords={'tile': ('sequence', tile_list), 'x': ('sequence', x_list), 'y': ('sequence', y_list)})
    return ds


def fastq_generator(fastq_filepath, chunksize):
    expr = re.compile('[:@ \n]')
    data = {'instrument': [], 'run': [], 'flowcell': [], 'lane': [],
            'tile': [], 'x': [], 'y': [], 'sequence': [], 'quality': []} # Perhaps better to use a numpy struct array
    current_chunk_size = 0
    with Path(fastq_filepath).open('r') as fq_file:
        for line_index, line in enumerate(fq_file):
            if line_index % 4 == 0:
                name = line.strip()
                instrument, run, flowcell, lane, tile, x, y = expr.split(name)[1:8]
                data['instrument'].append(instrument)
                data['run'].append(run)
                data['flowcell'].append(flowcell)
                data['lane'].append(lane)

                # sequence_index = numbered_index.loc[dict(sequence=(int(tile), int(x), int(y)))].item()
                data['tile'].append(int(tile))
                data['x'].append(int(x))
                data['y'].append(int(y))

            if line_index % 4 == 1:
                data['sequence'].append(line.strip())
            if line_index % 4 == 3:
                data['quality'].append(line.strip())
                # yield instrument_name, run_id, flowcell_id, lane, tile, x, y, sequence, quality

                current_chunk_size += 1

            if current_chunk_size == chunksize:
                yield data
                data = {'instrument': [], 'run': [], 'flowcell': [], 'lane': [],
                        'tile': [], 'x': [], 'y': [], 'sequence': [], 'quality': []}
                current_chunk_size = 0

        yield data # Final round

# def add_sequence_data_to_dataset(nc_filepath, fastq_filepath, name):
#     with xr.open_dataset(nc_filepath, engine='h5netcdf') as ds:
#         sequence_multiindex = ds[['tile', 'x', 'y']].load().set_index({'sequence':('tile','x','y')})#.indexes['sequence']
#
#     ds = fastq_data(fastq_filepath)
#     ds = ds.rename_vars({'read_sequence':f'{name}_sequence', 'read_quality': f'{name}_quality'})
#     ds = ds.set_index({'sequence': ('tile','x','y')})
#     ds, = xr.align(ds, indexes=sequence_multiindex.indexes, copy=False)
#     ds.reset_index('sequence',drop=True)[[f'{name}_sequence', f'{name}_quality']].to_netcdf(nc_filepath, mode='a', engine='h5netcdf')


def create_string_variable_in_nc_file(nc_file, variable_name, size=None, **kwargs):
    kwargs['dimensions'] += (variable_name + '_size',)
    # if not 'chunksizes' in kwargs.keys():
    #     kwargs['chunksizes'] = (10000, 1)
    nc_file.createDimension(variable_name + '_size', size)
    nc_file.createVariable(variable_name, 'S1', **kwargs)
    # nc_file[variable_name][:] = np.repeat(fill_value*size, nc_file.dimensions[kwargs['dimensions'][0]].size).astype(f'S{dtype_size}')
    # if fill_value is not None:
    #     nc_file[variable_name][:, :] = fill_value.encode()
    nc_file[variable_name]._Encoding = 'utf-8'



def add_sequence_data_to_dataset(nc_filepath, fastq_filepath, read_name, chunksize=10000):
    with netCDF4.Dataset(nc_filepath, 'a') as nc_file:
        nc_file['lane'][-1] = nc_file['lane'][-1]
        # To prevent that unlimited sequence dim is resized after reopening
        # Can probably be removed once libnetcdf 4.9 can be used.

    # import h5netcdf.legacyapi as h5netcdf
    # with h5netcdf.Dataset(nc_filepath, 'a') as nc_file:
        end_index = 0
        for i, sequence_data in tqdm.tqdm(enumerate(fastq_generator(fastq_filepath, chunksize=chunksize)), f'Add {read_name} to sequencing dataset', total=len(nc_file.dimensions['sequence'])//chunksize+1):
            if i == 0:
                size = len(sequence_data['sequence'][0])
                sequence_variable_name = read_name + '_sequence'
                create_string_variable_in_nc_file(nc_file, sequence_variable_name, dimensions=('sequence',), size=size)

                quality_variable_name = read_name + '_quality'
                create_string_variable_in_nc_file(nc_file, quality_variable_name, dimensions=('sequence',), size=size)

            start_index = end_index
            end_index = start_index + len(sequence_data['instrument'])
            # sequence_slice = slice(i * chunksize, (i + 1) * len(sequence_data))
            sequence_slice = slice(start_index, end_index)

            if (nc_file['tile'][sequence_slice] == sequence_data['tile']).all() & \
                    (nc_file['x'][sequence_slice] == sequence_data['x']).all() & \
                    (nc_file['y'][sequence_slice] == sequence_data['y']).all():
                nc_file[sequence_variable_name][sequence_slice, :] = np.array(sequence_data['sequence']).astype('S')
                nc_file[quality_variable_name][sequence_slice, :] = np.array(sequence_data['quality']).astype('S')
                pass
            else:
                raise ValueError()
#
#
# import xarray as xr
# from tqdm import tqdm
# nc_filepath = sam_filepath.with_suffix('.nc')
# with xr.open_dataset(nc_filepath, engine='h5netcdf') as ds:
#     sequence_multiindex = ds[['tile_number','x','y']].set_index({'sequence':('tile_number','x','y')}).indexes['sequence']
# numbered_index = xr.DataArray(np.arange(len(sequence_multiindex)), dims=('sequence',), coords={'sequence': sequence_multiindex})
# #
# # reference_array = ds[['tile','x','y']].to_array().T
# #
# # index1_sequence = xr.DataArray(np.empty(len(ds.sequence), dtype=str), dims=('sequence',), coords={'sequence': ds.sequence})
# # index1_quality = xr.DataArray(np.empty(len(ds.sequence), dtype=str), dims=('sequence',), coords={'sequence': ds.sequence})
# fastq_filepath = r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20211011 - Sequencer (MiSeq)\Analysis\Index1.fastq'
# with h5netcdf.File(nc_filepath, 'a') as nc_file:
#     if not 'index1_sequence' in nc_file:
#         nc_file.create_variable('index1_sequence', ('sequence',),  h5py.string_dtype(encoding='utf-8'))
#     if not 'index1_quality' in nc_file:
#         nc_file.create_variable('index1_quality', ('sequence',),  h5py.string_dtype(encoding='utf-8'))
#     sequence_index = 0
#     with Path(fastq_filepath).open('r') as fq_file:
#         for line_index, line in enumerate(tqdm(fq_file)):
#             if line_index % 4 == 0:
#                 name = line.strip()
#                 instrument_name, run_id, flowcell_id, lane, tile, x, y = re.split('[:@ \n]', name)[1:8]
#                 #sequence_index = numbered_index.loc[dict(sequence=(int(tile), int(x), int(y)))].item()
#
#             if line_index % 4 == 1:
#                 nc_file['index1_sequence'][sequence_index] = line.strip()
#             if line_index % 4 == 3:
#                 nc_file['index1_quality'][sequence_index] = line.strip()
#                 sequence_index += 1





        #index1_sequence.loc[dict(sequence=(int(tile), int(x), int(y)))] = fq_file.readline().strip()
        #fq_file.readline()
        #index1_quality.loc[dict(sequence=(int(tile), int(x), int(y)))] = fq_file.readline().strip()

def get_aligned_sequence(read_sequence, read_quality, cigar_string, position, reference_range=None):
    if reference_range is None:
        reference_range = (0, len(read_sequence))
    output_length = reference_range[1]-reference_range[0]

    if cigar_string == '*':
        return ('-'*output_length,' '*output_length)

    # cigar_string.split('(?<=M|I|D|N|S|H|P|=|X)')
    # split_cigar_string = re.findall(r'[0-9]*[MIDNSHP=X]', cigar_string)
    cigar_string_split = [(int(s[:-1]), s[-1]) for s in re.findall(r'[0-9]*[MIDNSHP=X]', cigar_string)]
    read_index = 0 # in read_sequence
    aligned_sequence = ''
    aligned_quality = ''

    if cigar_string_split[0][1] == 'S':
        read_index += cigar_string_split[0][0]
        cigar_string_split.pop(0)

    aligned_sequence += '-' * (position - 1)
    aligned_quality += ' ' * (position - 1)
    # read_index = position - 1

    for length, code in cigar_string_split:
        if code in ['M','=','X']:
            aligned_sequence += read_sequence[read_index:read_index+length]
            aligned_quality += read_quality[read_index:read_index+length]
            read_index += length
        elif code == 'I':
            read_index += length
        elif code == 'S':
            aligned_sequence += '-'*length
            aligned_quality += ' ' * length
            read_index += length
        elif code in ['D','N']:
            aligned_sequence += '-' * length
            aligned_quality += ' ' * length

    length_difference = len(aligned_sequence) - output_length
    if length_difference > 0:
        aligned_sequence = aligned_sequence[:output_length]
        aligned_quality = aligned_quality[:output_length]
    elif length_difference < 0:
        aligned_sequence += (-length_difference) * '-'
        aligned_quality += (-length_difference) * ' '

    return aligned_sequence, aligned_quality

def get_aligned_sequence_from_row(df_row, read_name, reference_range=None):
    sequence = df_row[read_name+'_sequence']
    quality = df_row[read_name+'_quality']
    cigar_string = df_row['cigar_string']
    position = df_row['position']
    return get_aligned_sequence(sequence, quality, cigar_string, position, reference_range)

########### Code to dynamically get aligned bases froj
# def get_aligned_position(index, cigar_string, first_base_position):
#     if cigar_string == '*':
#         return -1
#
#     # cigar_string.split('(?<=M|I|D|N|S|H|P|=|X)')
#     # split_cigar_string = re.findall(r'[0-9]*[MIDNSHP=X]', cigar_string)
#     cigar_string_split = [(int(s[:-1]), s[-1]) for s in re.findall(r'[0-9]*[MIDNSHP=X]', cigar_string)]
#     query_index = 0 # in read_sequence
#     reference_index = first_base_position-1
#     for length, code in cigar_string_split:
#         if code in ['M','I','S','=','X']:
#             query_index += length
#         if code in ['M','D','N','=','X']:
#             reference_index += length
#         #print(f'Reference index: {reference_index}')
#         #print(f'Query index: {query_index}')
#         if reference_index > index:
#             return query_index-reference_index+index
#     return -1
#
# def get_aligned_position_from_row(df_row, index):
#     cigar_string = df_row['cigar_string']
#     first_base_position = df_row['first_base_position']
#     return get_aligned_position(index, cigar_string, first_base_position)
#
#     a = np.array(re.findall(r'[0-9]*(?=[MIS=X])', cigar_string)).astype(int).cumsum()
#     b = np.array([first_base_position-1]+re.findall(r'[0-9]*(?=[MDN=X])', cigar_string)).astype(int).cumsum()
#     return aligned_sequence, aligned_quality


def extract_positions(series, indices):
    for i, index in enumerate(indices):
        if i == 0:
            combined = series.str.get(index)
        else:
            combined += series.str.get(index)
    return combined


def make_sequencing_dataset(file_path, index1_file_path=None, remove_duplicates=True, add_aligned_sequence=True,
                            extract_sequence_subset=False, chunksize=10000):

    file_path = Path(file_path)

    nc_file_path = file_path.with_name('sequencing_data.nc')
    if file_path.suffix == '.sam':
        parse_sam(file_path, remove_duplicates=remove_duplicates, add_aligned_sequence=add_aligned_sequence,
                  extract_sequence_subset=extract_sequence_subset, chunksize=chunksize, write_csv=False, write_nc=True,
                  write_filepath=nc_file_path)
    else:
        raise ValueError('Wrong file type')

    if index1_file_path is not None:
        index1_file_path = Path(index1_file_path)
        if index1_file_path.suffix == '.fastq':
            add_sequence_data_to_dataset(nc_file_path, index1_file_path, 'index1')
        else:
            raise ValueError('Wrong file type for index1')

    return nc_file_path



# TODO: Add to sam to nc file conversion
import re
import tqdm

def determine_read_end_position(cigar_string, reference_start_postion, reference_length):
    cigar_string_split = [(int(s[:-1]), s[-1]) for s in re.findall(r'[0-9]*[MIDNSHP=X]', cigar_string)]
    reference_index = reference_start_postion
    read_index = 0
    for length, code in cigar_string_split:
        if code in ['M',  'D', 'N', '=', 'X']:
            reference_index += length
        if code in ['M', 'I', 'S', '=', 'X']:
            read_index += length
        if reference_index >= reference_length:
            return read_index-(reference_index-reference_length)+1
    return reference_length

def determine_read_end_positions(sequencing_data, reference_length, chunk_size=10000):
    end_positions = np.zeros(len(sequencing_data)).astype('int16')
    for j in tqdm.tqdm(list(range(0,len(sequencing_data),chunk_size))):
        ds = sequencing_data.dataset.isel(sequence=slice(j, j+chunk_size))
        cigar_strings = ds.cigar_string.load()
        reference_start_positions = ds.position.load()
        for i, (cigar_string, reference_start_position) in enumerate(zip(cigar_strings, reference_start_positions)):
            end_positions[j+i] = determine_read_end_position(cigar_string.item(), reference_start_position.item(), reference_length)
    return end_positions


def sequence_correspondence(sequences, references, minimum_match=None):
    reference_names = np.array(list(references.keys()))
    reference_barcodes = np.array(list(references.values()))
    length = len(sequences[0])
    sequences_split = sequences.astype('S').view('S1').reshape(-1,length)
    reference_barcodes_split = reference_barcodes.astype('S').view('S1').reshape(-1,length)
    N_matched = (sequences_split[:, None, :] == reference_barcodes_split[None, :, :]).sum(axis=-1)
    correspondence = reference_names[N_matched.argmax(axis=1)]
    score = N_matched.max(axis=1)
    if minimum_match is not None:
        correspondence[score < minimum_match] = ''
    return correspondence, score

#
# def write_string_variable_to_nc_file(nc_file, name, array, indices=None):
#     from papylio.plugins.sequencing.sequencing_data import create_string_variable_in_nc_file
#
#     if indices is None:
#         indices = slice(None)
#
#     array = array.astype('S')
#     size = int(str(array.dtype)[2:])
#     create_string_variable_in_nc_file(nc_file, name, dimensions=('sequence',), size=size)
#     nc_file[name][indices] = array

if __name__ == '__main__':
    file_path = r'J:\Ivo\20211011 - Sequencer (MiSeq)\Analysis\sequencing_data_MapSeq.csv'
    seqdata = SequencingData(file_path=file_path)
    seqdata.plot_cluster_locations_per_tile(save_filepath=r'J:\Ivo\20211011 - Sequencer (MiSeq)\Analysis\Mapping_seqquences_per_tile.png')

    file_path = r'J:\Ivo\20211011 - Sequencer (MiSeq)\Analysis\sequencing_data_HJ_general.csv'
    seqdata_HJ = SequencingData(file_path=file_path)



    # New analysis
    sam_filepath = Path(r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20211011 - Sequencer (MiSeq)\Analysis\Alignment.sam')
    # sequencing_data = read_sam(sam_filepath)

    extract_sequence_subset = [30, 31, 56, 57, 82, 83, 108, 109]
    parse_sam(sam_filepath, remove_duplicates=True, add_aligned_sequence=True,
              extract_sequence_subset=extract_sequence_subset,
              chunksize=10000, write_csv=False, write_nc=True)

    nc_filepath = sam_filepath.with_suffix('.nc')
    fastq_filepath = r'N:\tnw\BN\CMJ\Shared\Ivo\PhD_data\20211011 - Sequencer (MiSeq)\Analysis\Index1.fastq'
    add_sequence_data_to_dataset(nc_filepath, fastq_filepath, 'index1')



# -*- coding: utf-8 -*-
"""
Created on Tue Oct 15 11:09:03 2019

@author: pimam
"""
if __name__ == '__main__':
    import os
    import sys
    from pathlib import Path, PureWindowsPath
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    p = Path(__file__).parents[2]
    sys.path.insert(0, str(p))
    main_path = PureWindowsPath('F:\\20191211_dCas9_DNA5_7_8_20_Vikttracr\\#5_strept_1nMCas9_10nMDNA20_snaps\\peaks5collect4_green_red\\histograms')

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from papylio.experiment import Experiment
import seaborn as sns
sns.set(style="ticks")
sns.set_color_codes()

def plot_avgFRET(exp, bins=100, save=True):
    
    Edata=[]
    for fl in exp.files:
        nmole= len(fl.molecules)
        print(nmole)
        data = np.array([np.mean(molecule.E()) for molecule in fl.molecules])
        Edata = np.concatenate((Edata, data), axis=0)
        plt.figure()
        plt.hist(data, bins, range = (0,1))
        plt.title(f'avgFRET {fl.name} bins:{bins} mole:{nmole}') 
        plt.xlabel('FRET')
        plt.ylabel('count')
        plt.savefig(f'{fl.name}_avgFRET_hist.png', facecolor='white', dpi=200)
    nmole = np.size(Edata)
    plt.figure()
    plt.hist(data, bins, range = (0,1))
    plt.title(f'avgFRET all snaps bins:{bins} mole:{nmole}')
    plt.xlabel('FRET')
    plt.ylabel('count')
    plt.savefig('avgFRETall_hist.png', facecolor='white', dpi=300)

if __name__ == '__main__':
    exp = Experiment(main_path)
    bins=100
    plot_avgFRET(exp, bins, save=True)

